package es.accv.arangi.client;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

import com.sun.org.apache.bcel.internal.Constants;

import es.accv.arangi.base.certificate.Certificate;
import es.accv.arangi.base.certificate.validation.CertificateValidator;
import es.accv.arangi.base.certificate.validation.OCSPResponse;
import es.accv.arangi.base.certificate.validation.ValidateCertificate;
import es.accv.arangi.base.certificate.validation.ValidatingTrace;
import es.accv.arangi.base.device.AppleSafariManager;
import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.document.ByteArrayDocument;
import es.accv.arangi.base.document.InputStreamDocument;
import es.accv.arangi.base.exception.certificate.NormalizeCertificateException;
import es.accv.arangi.base.exception.device.AppleSafariKeyStoreException;
import es.accv.arangi.base.exception.device.AppleSafariNotFoundException;
import es.accv.arangi.base.exception.device.DeviceNotFoundException;
import es.accv.arangi.base.exception.device.IAIKDLLNotFoundException;
import es.accv.arangi.base.exception.device.IncorrectPINException;
import es.accv.arangi.base.exception.device.InitializeProviderException;
import es.accv.arangi.base.exception.device.LockedPINException;
import es.accv.arangi.base.exception.device.ModuleNotFoundException;
import es.accv.arangi.base.exception.device.MozillaKeyStoreException;
import es.accv.arangi.base.exception.device.MozillaNotFoundException;
import es.accv.arangi.base.exception.device.OpeningDeviceException;
import es.accv.arangi.base.exception.device.WindowsNotFoundException;
import es.accv.arangi.base.exception.signature.InvalidCertificateException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.exception.signature.SignatureNotFoundException;
import es.accv.arangi.base.exception.timestamp.MalformedTimeStampException;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.base.util.validation.ValidationResult;
import es.accv.arangi.certificate.CertificateFactory;
import es.accv.arangi.client.exception.MalformedCertificateException;
import es.accv.arangi.client.exception.NoArangiCertificateException;
import es.accv.arangi.client.exception.NoCertificateToSignException;
import es.accv.arangi.client.exception.NoDetachedDocumentToValidateException;
import es.accv.arangi.client.exception.SignatureCancelledByUserException;
import es.accv.arangi.client.exception.SignatureFormatException;
import es.accv.arangi.client.exception.SignatureTypeUnknownException;
import es.accv.arangi.client.filter.ArangiClientFilter;
import es.accv.arangi.client.model.SignerInformation;
import es.accv.arangi.client.model.ValidationObject;
import es.accv.arangi.client.signature.ArangiClientSignatureFactory;
import es.accv.arangi.client.util.DeviceAndAlias;
import es.accv.arangi.client.util.I18n;
import es.accv.arangi.client.validation.ArangiClientValidationFactory;
import es.accv.arangi.device.MozillaKeyStoreManager;
import es.accv.arangi.device.PersonalWindowsKeyStoreManager;
import es.accv.arangi.device.Pkcs11Manager;
import es.accv.arangi.timestamp.TimeStamp;

/**
 * Cliente de Arang
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class ArangiClient {

	/**
	 * Serial Version UID
	 */
	public static final long serialVersionUID = -1205867543798348959L;

	/*
	 * Paquete donde se encuentran las clases que realizarn las firmas
	 */
	private static final String SIGNATURE_PACKAGE = "es.accv.arangi.client.signature";

	/*
	 * Paquete donde se encuentran las clases que realizarn las validaciones
	 */
	private static final String VALIDATION_PACKAGE = "es.accv.arangi.client.validation";

	/**
	 * Constante para formatear fechas
	 */
	public static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("dd-MM-yyyy HH:mm");
	
	/**
	 * Opcin que se debe pasar a esta clase para que se firme sin sello de tiempos
	 */
	public static final String OPCION_FIRMAR_SIN_SELLO = "firmarSinSello";

	/**
	 * Opcin que se debe pasar a las clases XAdES para que se aada la descripcin del documento
	 */
	public static final String OPCION_DESCRIPTION 	= "description";

	/**
	 * Opcin que se debe pasar a las clases XAdES para que se aada el tipo MIME del documento
	 */
	public static final String OPCION_MIME 		= "mime";
	
	/**
	 * Opcin que se debe pasar a las clases XAdES para que se aada la codificacin del documento
	 */
	public static final String OPCION_ENCODING 	= "encoding";

	/**
	 * Opcin que se debe pasar a las clases PDF o PAdES para pasar una imagen como firma visible
	 */
	public static final String OPCION_IMAGE_PATH 	= "imagePath";

	/**
	 * Opcin que se debe pasar a las clases PDF o PAdES para describir la ubicacin de una imagen 
	 * en una firma visible
	 */
	public static final String OPCION_X_BOTTOM_LEFT = "xBottomLeft";
	public static final String OPCION_Y_BOTTOM_LEFT = "yBottomLeft";
	public static final String OPCION_X_TOP_RIGHT 	= "xTopRight";
	public static final String OPCION_Y_TOP_RIGHT 	= "yTopRight";

	/**
	 * Opcin que se debe pasar a las clases PDF o PAdES para indicar la pgina de la firma visible
	 */
	public static final String OPCION_PAGE 	= "page";

	/*
	 * Logger de la clase
	 */
	private static Logger logger = Logger.getLogger(ArangiClient.class);
	
	/*
	 * i18n
	 */
	private I18n i18n; // por defecto en castellano
	
	/*
	 * Flag que indica qu hacer si slo hay un certificado en la seleccin
	 */
	private boolean mostrarSeleccionSiSoloUno	= true;
	
	/*
	 * URL a la TSA
	 */
	private String urlTSA = TimeStamp.URL_ACCV_TSA_SERVER;
	
	/*
	 * Opciones que se pasarn en el momento de generar la firma
	 */
	protected Map<String, Object> options = new HashMap<String, Object>();
	
	//-- Constructor
	
	/**
	 * Constructor
	 */
	public ArangiClient(String language, String country) {
		
		//-- Cargar el bundle. Se busca el locale primero por los parmetros, segundo
		//-- por un parmetro de usuario del System (para JNLP) y tercero por el
		//-- parmetro de sistema del System
		Locale locale = null;
		if (language != null && country != null) {
			locale = new Locale(language, country);
		} 
		if (locale == null && System.getProperty("language") != null && System.getProperty("country") != null) {
			locale = new Locale(System.getProperty("language"), System.getProperty("country"));
		}
		if (locale == null) {
			locale = new Locale(System.getProperty("user.language"), System.getProperty("user.country"));
		}
		
		logger.debug("[ArangiClient]::Locale: " + locale);
		
		i18n = new I18n (locale); 
		
	}

	//-- Mtodos pblicos de firma
	
	/**
	 * Carga las opciones que se utilizarn durante la firma
	 */
	public void setOpciones (Map<String, Object> options) {
		this.options = options;
	}
	
	/**
	 * Carga la variable que indica si se muestra la seleccin si slo hay un
	 * elemento a seleccionar
	 */
	public void setMostrarSeleccionSiSoloUno(boolean mostrarSeleccionSiSoloUno) {
		this.mostrarSeleccionSiSoloUno = mostrarSeleccionSiSoloUno;
	}
	
	/**
	 * Cambia el valor de la TSA a utilizar
	 */
	public void setUrlTSA (String urlTSA) {
		this.urlTSA = urlTSA;
	}
	
	/**
	 * Firma un fichero que ser elegido por el usuario.
	 * 
	 * @param filter Filtro para elegir los certificados
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero elegido.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws NoDocumentToSignException El usuario no desea elegir un documento
	 * 	para firmar
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado de firma
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la firma
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 */
	public String signFile (ArangiClientFilter filter, String reference, String signatureType) throws NoDocumentToSignException, OpeningDeviceException, SignatureTypeUnknownException, SignatureException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.signFile]::Entrada::" + Arrays.asList(new Object[] { filter, reference, signatureType }));
		
		//-- Mostrar dilogo para obtener el path al fichero a firmar
		JFileChooser jfc = new JFileChooser();
		jfc.setDialogTitle(i18n.getString("titulo.elegir.documento.firmar"));
		int ret = jfc.showOpenDialog(null);
		if (ret == JFileChooser.APPROVE_OPTION) {
			String path = jfc.getSelectedFile().getAbsolutePath();
			logger.debug("[ArangiClient.signFile]::Documento a firmar: " + path);
			return signFile(new File (path), filter, reference, signatureType);
		}
		
		//-- El usuario ha pulsado el botn de cancelar
		logger.warn("[ArangiClient.signFile]::El usuario no ha elegido un documento para firmar");
		throw new NoDocumentToSignException(i18n.getString("error.no.documento.firmar"));
	}
	
	
	/**
	 * Firma el fichero que se pasa como parmetro.
	 * 
	 * @param file Fichero que hay que firmar
	 * @param filter Filtro para elegir los certificados
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado de firma
	 * @throws NoDocumentToSignException El fichero no existe o no se puede leer
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la firma
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 */
	public String signFile (File file, ArangiClientFilter filter, String reference, String signatureType) throws OpeningDeviceException, NoDocumentToSignException, SignatureTypeUnknownException, SignatureException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.signFile]::Entrada::" + Arrays.asList(new Object[] { file, filter, reference, signatureType }));
		
		//-- Elegir el certificado de firma
		CertificateSelector selector = new CertificateSelector(i18n, filter, mostrarSeleccionSiSoloUno);
		DeviceAndAlias element = selector.getDeviceAndAlias();
		
		//-- Firmar	
		logger.info("[ArangiClient.signFile]::Firmar con certificado: " + element.getCertificate().getCommonName());
		return signFile (file, element, reference, signatureType);
	}
	
	/**
	 * Firma el fichero que se pasa como parmetro con el elemento.
	 * 
	 * @param file Fichero que hay que firmar
	 * @param signCertificate Informacin del keystore y su certificado para poder
	 * firmar
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado de firma
	 * @throws NoDocumentToSignException El fichero no existe o no se puede leer
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la firma
	 */
	public String signFile(File file, DeviceAndAlias signCertificate, String reference, String signatureType) throws OpeningDeviceException, NoDocumentToSignException, SignatureTypeUnknownException, SignatureException {
		logger.info("[ArangiClient.signFile]::Entrada::" + Arrays.asList(new Object[] { file, signCertificate.getCertificate().getCommonName(), reference, signatureType }));
		
		DeviceManager manager = null;
		try {
			//-- Obtener el manager
			manager = getManager (signCertificate);
			logger.debug("[ArangiClient.signFile]::Obtenido el manager");
			
			//-- Firmar en el formato adecuado
			return signFile (file, manager, signCertificate.getAlias(), reference, signatureType);
			
		} finally {
			//-- Siempre cerrar el manager
			if (manager != null) {
				manager.close();
			}
		}

	}

	/**
	 * Firma el fichero con el alias contenido en el manager que se pasa como parmetro.
	 * 
	 * @param file Fichero que hay que firmar
	 * @param manager Dispositivo abierto
	 * @param alias Alias del certificado dentro del dispositivo
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws NoDocumentToSignException El fichero no existe o no se puede leer
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la firma
	 */
	public String signFile(File file, DeviceManager manager, String alias, String reference, String signatureType) throws NoDocumentToSignException, SignatureTypeUnknownException, SignatureException {
		logger.info("[ArangiClient.signFile]::Entrada::" + Arrays.asList(new Object[] { file, manager, alias, reference, signatureType }));
		
		//-- Obtener el objeto que obtendr la firma
		ArangiClientSignatureFactory clientSignature = getClientSignatureFactory (signatureType);
		logger.debug ("[ArangiClient.signFile]::Se ha obtenido el objeto que realizar la firma: " + clientSignature);
		
		//-- Firmar
		logger.debug ("[ArangiClient.signFile]::Firmando");	
		clientSignature.setSignatureOptions(this.options);
		clientSignature.setUrlTSA(this.urlTSA);
		return clientSignature.sign (file, manager, alias, reference);
	}

	/**
	 * Firma el contenido que se pasa como parmetro.
	 * 
	 * @param contentToSign Contenido a firmar
	 * @param filter Filtro para elegir los certificados
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs
	 * @return Firma en base64
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado de firma
	 * @throws NoDocumentToSignException El fichero no existe o no se puede leer
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la firma
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 */
	public String signContent (byte[] contentToSign, ArangiClientFilter filter, String reference, String signatureType) throws OpeningDeviceException, NoDocumentToSignException, SignatureTypeUnknownException, SignatureException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.signContent]::Entrada::" + Arrays.asList(new Object[] { contentToSign, filter, signatureType, reference }));
		
		//-- Elegir el certificado de firma
		CertificateSelector selector = new CertificateSelector(i18n, filter, mostrarSeleccionSiSoloUno);
		DeviceAndAlias element = selector.getDeviceAndAlias();
		
		//-- Firmar	
		logger.info("[ArangiClient.signContent]::Firmar con certificado: " + element.getCertificate().getCommonName());
		return signContent (contentToSign, element, reference, signatureType);
	}
	
	/**
	 * Firma el contenido que se pasa como parmetro con el elemento.
	 * 
	 * @param contentToSign Contenido a firmar
	 * @param signCertificate Informacin del keystore y su certificado para poder
	 * firmar
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado de firma
	 * @throws NoDocumentToSignException El fichero no existe o no se puede leer
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la firma
	 */
	public String signContent(byte[] contentToSign, DeviceAndAlias signCertificate, String reference, String signatureType) throws OpeningDeviceException, NoDocumentToSignException, SignatureTypeUnknownException, SignatureException {
		logger.info("[ArangiClient.signContent]::Entrada::" + Arrays.asList(new Object[] { contentToSign, signCertificate.getCertificate().getCommonName(), signatureType, reference }));
		
		DeviceManager manager = null;
		try {
			//-- Obtener el manager
			manager = getManager (signCertificate);
			logger.debug("[ArangiClient.signContent]::Obtenido el manager");
			
			//-- Firmar en el formato adecuado
			return signContent (contentToSign, manager, signCertificate.getAlias(), reference, signatureType);
			
		} finally {
			//-- Siempre cerrar el manager
			if (manager != null) {
				manager.close();
			}
		}

	}

	/**
	 * Firma el contenido con el alias contenido en el manager que se pasa como parmetro.
	 * 
	 * @param contentToSign Contenido a firmar
	 * @param manager Dispositivo abierto
	 * @param alias Alias del certificado dentro del dispositivo
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Firma en base64
	 * @throws NoDocumentToSignException El fichero no existe o no se puede leer
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la firma
	 */
	public String signContent(byte[] contentToSign, DeviceManager manager, String alias, String reference, String signatureType) throws NoDocumentToSignException, SignatureTypeUnknownException, SignatureException {
		logger.info("[ArangiClient.signContent]::Entrada::" + Arrays.asList(new Object[] { contentToSign, manager, alias, signatureType }));
		
		//-- Obtener el objeto que obtendr la firma
		ArangiClientSignatureFactory signatureFactory = getClientSignatureFactory (signatureType);
		logger.debug ("[ArangiClient.signContent]::Se ha obtenido el objeto que realizar la firma: " + signatureFactory);
		
		//-- Firmar
		logger.debug ("[ArangiClient.signContent]::Firmando");		
		signatureFactory.setSignatureOptions(this.options);
		signatureFactory.setUrlTSA(this.urlTSA);
		return signatureFactory.sign (contentToSign, manager, alias, reference);
	}
	
	/**
	 * El usuario elige un fichero con la firma y el documento que la origin y con ellos
	 * se realiza una cofirma.<br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param filter Filtro para elegir los certificados
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero elegido.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureNotFoundException El fichero de firma no existe o no puede ser ledo
	 * @throws SignatureException Error creando la cofirma
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para cofirmar
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 */
	public String coSign (ArangiClientFilter filter, String reference, String signatureType) throws SignatureException, SignatureNotFoundException, SignatureTypeUnknownException, OpeningDeviceException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { filter, reference, signatureType }));
		
		//-- Mostrar dilogo para obtener el path al fichero con la firma
		JFileChooser jfc = new JFileChooser();
		jfc.setDialogTitle(i18n.getString("titulo.elegir.firma"));
		int ret = jfc.showOpenDialog(null);
		if (ret != JFileChooser.APPROVE_OPTION) {
			//-- El usuario ha pulsado el botn de cancelar
			logger.warn("[ArangiClient.coSign]::El usuario no ha elegido un documento para firmar");
			throw new SignatureNotFoundException(i18n.getString("error.no.documento.firmar"));
		}
	
		String path = jfc.getSelectedFile().getAbsolutePath();
		logger.debug("[ArangiClient.coSign]::Firma: " + path);
		File signature = new File (path);
		
		//-- Mostrar dilogo para obtener el path al documento que origin la firma
		jfc = new JFileChooser();
		jfc.setDialogTitle(i18n.getString("titulo.elegir.documento.firma"));
		ret = jfc.showOpenDialog(null);
		if (ret == JFileChooser.APPROVE_OPTION) {
			path = jfc.getSelectedFile().getAbsolutePath();
			logger.debug("[ArangiClient.coSign]::Documento que origin la firma: " + path);
			return coSign (signature, new File (path), filter, reference, signatureType);
		} else {
			logger.debug("[ArangiClient.coSign]::No se pasa el documento que origin la firma y el tipo de firma es : " + signatureType);
			return coSign (signature, (File) null, filter, reference, signatureType);
		}
		
	}
	
	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro.<br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signature Fichero con la firma
	 * @param document Documento que origin la firma
	 * @param filter Filtro para elegir los certificados
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureNotFoundException El fichero de firma no existe o no puede ser ledo
	 * @throws SignatureException Error creando la cofirma
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para cofirmar
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 */
	public String coSign (File signature, File document, ArangiClientFilter filter, String reference, String signatureType) throws SignatureException, SignatureNotFoundException, SignatureTypeUnknownException, OpeningDeviceException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signature, document, filter, reference, signatureType }));
		
		//-- Elegir el certificado de firma
		CertificateSelector selector = new CertificateSelector(i18n, filter, mostrarSeleccionSiSoloUno);
		DeviceAndAlias element = selector.getDeviceAndAlias();
		
		//-- Cofirmar	
		logger.info("[ArangiClient.coSign]::Cofirmar con certificado: " + element.getCertificate().getCommonName());
		return coSign (signature, document, element, reference, signatureType);
	}
	
	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro.<br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signature Fichero con la firma
	 * @param document Contenido del documento que origin la firma
	 * @param filter Filtro para elegir los certificados
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureNotFoundException El fichero de firma no existe o no puede ser ledo
	 * @throws SignatureException Error creando la cofirma
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para cofirmar
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 */
	public String coSign (File signature, byte[] document, ArangiClientFilter filter, String reference, String signatureType) throws SignatureException, SignatureNotFoundException, SignatureTypeUnknownException, OpeningDeviceException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signature, document, filter, reference, signatureType }));
		
		//-- Contenido de la firma
		byte[] signatureContent;
		try {
			signatureContent = Util.loadFile(signature);
		} catch (IOException e) {
			logger.error("[ArangiClient.coSign]::No existe el fichero con la firma " + signature, e);
			throw new SignatureNotFoundException(i18n.getString("error.no.firma", signature), e);
		}

		//-- Cofirmar	
		logger.info("[ArangiClient.coSign]::Cofirmar");
		return coSign (signatureContent, document, filter, reference, signatureType);
	}
	
	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro.<br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signature Contenido del fichero con la firma
	 * @param document Documento que origin la firma
	 * @param filter Filtro para elegir los certificados
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureNotFoundException El fichero de firma no existe o no puede ser ledo
	 * @throws SignatureException Error creando la cofirma
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para cofirmar
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 */
	public String coSign (byte[] signature, File document, ArangiClientFilter filter, String reference, String signatureType) throws SignatureException, SignatureNotFoundException, SignatureTypeUnknownException, OpeningDeviceException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signature, document, filter, reference, signatureType }));
		
		//-- Contenido de la firma
		byte[] documentContent = null;
		try {
			documentContent = Util.loadFile(document);
		} catch (IOException e) {
		}

		//-- Cofirmar	
		logger.info("[ArangiClient.coSign]::Cofirmar");
		return coSign (signature, documentContent, filter, reference, signatureType);
	}
	
	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro. La cofirma se
	 * lleva a cabo con el certificado pasado como parmetro.<br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signature Fichero con la firma
	 * @param document Documento que origin la firma
	 * @param signCertificate Informacin del keystore y su certificado para poder
	 * firmar
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureNotFoundException El fichero de firma no existe o no puede ser ledo
	 * @throws SignatureException Error creando la cofirma
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para cofirmar
	 */
	public String coSign(File signature, File document, DeviceAndAlias signCertificate, String reference, String signatureType) throws SignatureException, SignatureNotFoundException, SignatureTypeUnknownException, OpeningDeviceException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signature, document, signCertificate.getCertificate().getCommonName(), reference, signatureType }));
		
		DeviceManager manager = null;
		try {
			//-- Obtener el manager
			manager = getManager (signCertificate);
			logger.debug("[ArangiClient.signFile]::Obtenido el manager");
			
			//-- Cofirmar en el formato adecuado
			return coSign (signature, document, manager, signCertificate.getAlias(), reference, signatureType);
			
		} finally {
			//-- Siempre cerrar el manager
			if (manager != null) {
				manager.close();
			}
		}

	}

	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro. La cofirma se
	 * lleva a cabo con el alias contenido en el manager que se pasa como parmetro.<br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signature Fichero con la firma
	 * @param document Documento que origin la firma
	 * @param manager Dispositivo abierto
	 * @param alias Alias del certificado dentro del dispositivo
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero.
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureNotFoundException El fichero de firma no existe o no puede ser ledo
	 * @throws SignatureException Error creando la cofirma
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 */
	public String coSign(File signature, File document, DeviceManager manager, String alias, String reference, String signatureType) throws SignatureException, SignatureNotFoundException, SignatureTypeUnknownException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signature, document, manager, alias, reference, signatureType }));
		
		//-- Obtener el objeto que obtendr la firma
		ArangiClientSignatureFactory clientSignature = getClientSignatureFactory (signatureType);
		logger.debug ("[ArangiClient.coSign]::Se ha obtenido el objeto que realizar la firma: " + clientSignature);
		
		//-- Cofirmar
		logger.debug ("[ArangiClient.coSign]::Cofirmando");		
		clientSignature.setSignatureOptions(this.options);
		clientSignature.setUrlTSA(this.urlTSA);
		return clientSignature.coSign (signature, document, manager, alias, reference);
	}

	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro. <br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signatureContent Contenido de la firma
	 * @param documentContent Contenido del documento que origin la firma
	 * @param filter Filtro para elegir los certificados
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs
	 * @return Firma en base64
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la cofirma
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para cofirmar
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 */
	public String coSign (byte[] signatureContent, byte[] documentContent, ArangiClientFilter filter, String reference, String signatureType) throws SignatureTypeUnknownException, SignatureException, OpeningDeviceException, SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signatureContent, documentContent, filter, signatureType, reference }));
		
		//-- Elegir el certificado de firma
		CertificateSelector selector = new CertificateSelector(i18n, filter, mostrarSeleccionSiSoloUno);
		DeviceAndAlias element = selector.getDeviceAndAlias();
		
		//-- Firmar	
		logger.info("[ArangiClient.coSign]::Firmar con certificado: " + element.getCertificate().getCommonName());
		return coSign (signatureContent, documentContent, element, reference, signatureType);
	}
	
	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro. <br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signatureContent Contenido de la firma
	 * @param documentContent Contenido del documento que origin la firma
	 * @param signCertificate Informacin del keystore y su certificado para poder
	 * firmar
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs
	 * @param signatureType Tipo de firma segn las constantes definidas en
	 * 	la clase {@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la cofirma
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para cofirmar
	 */
	public String coSign(byte[] signatureContent, byte[] documentContent, DeviceAndAlias signCertificate, String reference, String signatureType) throws SignatureTypeUnknownException, SignatureException, OpeningDeviceException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signatureContent, documentContent, signCertificate.getCertificate().getCommonName(), signatureType, reference }));
		
		DeviceManager manager = null;
		try {
			//-- Obtener el manager
			manager = getManager (signCertificate);
			logger.debug("[ArangiClient.coSign]::Obtenido el manager");
			
			//-- Cofirmar en el formato adecuado
			return coSign (signatureContent, documentContent, manager, signCertificate.getAlias(), reference, signatureType);
			
		} finally {
			//-- Siempre cerrar el manager
			if (manager != null) {
				manager.close();
			}
		}

	}

	/**
	 * Realiza la cofirma del documento en la firma pasada como parmetro con el alias 
	 * contenido en el manager que se pasa como parmetro. <br><br>
	 * 
	 * Si el documento pasado no es el mismo que origin la primera firma se producir una
	 * excepcin.
	 * 
	 * @param signatureContent Contenido de la firma
	 * @param documentContent Contenido del documento que origin la firma
	 * @param manager Dispositivo abierto
	 * @param alias Alias del certificado dentro del dispositivo
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Firma en base64
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la cofirma
	 */
	public String coSign(byte[] signatureContent, byte[] documentContent, DeviceManager manager, String alias, String reference, String signatureType) throws SignatureTypeUnknownException, SignatureException {
		logger.info("[ArangiClient.coSign]::Entrada::" + Arrays.asList(new Object[] { signatureContent, documentContent, manager, alias, signatureType }));
		
		//-- Obtener el objeto que obtendr la firma
		ArangiClientSignatureFactory signatureFactory = getClientSignatureFactory (signatureType);
		logger.debug ("[ArangiClient.coSign]::Se ha obtenido el objeto que realizar la cofirma: " + signatureFactory);
		
		//-- Firmar
		logger.debug ("[ArangiClient.coSign]::Cofirmando");		
		signatureFactory.setSignatureOptions(this.options);
		signatureFactory.setUrlTSA(this.urlTSA);
		return signatureFactory.coSign (signatureContent, documentContent, manager, alias, reference);
	}
	
	/**
	 * Slo realiza contrafirma para los tipos XAdES.<br><br>
	 * 
	 * La definicin de las contrafirmas en XAdES puede observarse en el punto 7.2.4
	 * del estndar de la ETSI.<br><br>
	 * 
	 * Este mtodo realiza una contrafirma para la ltima firma del XAdES permitiendo
	 * al usuario elegir el certificado para contrafirmar. Es til cuando se sabe que 
	 * el XAdES contiene slo una firma.<br><br>
	 * 
	 * @param signature Firma (XAdES)
	 * @param filter Filtro para elegir los certificados
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Firma actualizada con la contrafirma en base64
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la contrafirma
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para contrafirmar
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 */
	public String counterSign (byte[] signature, ArangiClientFilter filter, String signatureType) throws SignatureTypeUnknownException, SignatureException, SignatureCancelledByUserException, NoCertificateToSignException, OpeningDeviceException, SignatureFormatException {
		logger.info("[ArangiClient.counterSign]::Entrada::" + Arrays.asList(new Object[] { signature, signatureType }));
		
		//-- Obtener el objeto que obtendr la firma
		ArangiClientSignatureFactory signatureFactory = getClientSignatureFactory (signatureType);
		logger.debug ("[ArangiClient.counterSign]::Se ha obtenido el objeto que realizar la contrafirma: " + signatureFactory);
		
		//-- Elegir la firma que se quiere contrafirmar dentro del objeto firma
		SignatureSelector signatureSelector;
		try {
			signatureSelector = new SignatureSelector (i18n, signatureFactory.getSignatureObject(signature));
		} catch (Exception e) {
			logger.error("[ArangiClient.counterSign]::No se ha podido obtener el objeto firma del tipo '" + signatureType + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", signatureType), e);
		}
		Certificate signatureToCounterSignCertificate = signatureSelector.getCertificateSignature ();
		
		//-- Elegir el certificado para realizar la contrafirma
		CertificateSelector certificateSelector = new CertificateSelector(i18n, filter, mostrarSeleccionSiSoloUno);
		DeviceAndAlias signatureCertificate = certificateSelector.getDeviceAndAlias();
		
		DeviceManager manager = null;
		try {
			//-- Obtener el manager
			manager = getManager (signatureCertificate);
			logger.debug("[ArangiClient.counterSign]::Obtenido el manager");
			
			//-- Contrafirmar
			signatureFactory.setSignatureOptions(this.options);
			signatureFactory.setUrlTSA(this.urlTSA);
			return signatureFactory.counterSign(signature, manager, signatureCertificate.getAlias(), signatureToCounterSignCertificate);
			
		} finally {
			//-- Siempre cerrar el manager
			if (manager != null) {
				manager.close();
			}
		}
	}

	/**
	 * Slo realiza contrafirma para los tipos XAdES.<br><br>
	 * 
	 * La definicin de las contrafirmas en XAdES puede observarse en el punto 7.2.4
	 * del estndar de la ETSI.<br><br>
	 * 
	 * Este mtodo realiza una contrafirma para la ltima firma del XAdES permitiendo
	 * al usuario elegir el certificado para contrafirmar. Es til cuando se sabe que 
	 * el XAdES contiene slo una firma.<br><br>
	 * 
	 * @param signatureFile Fichero con la firma (XAdES)
	 * @param filter Filtro para elegir los certificados
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Firma actualizada con la contrafirma en base64
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureException Error creando la contrafirma
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 * @throws SignatureCancelledByUserException El usuario no ha querido elegir un certificado
	 * 	de la lista
	 * @throws OpeningDeviceException No se ha podido abrir el dispositivo con el
	 * 	certificado para contrafirmar
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 */
	public String counterSign (File signatureFile, ArangiClientFilter filter, String signatureType) throws SignatureTypeUnknownException, SignatureException, SignatureCancelledByUserException, NoCertificateToSignException, OpeningDeviceException, SignatureFormatException {
		logger.info("[ArangiClient.counterSign]::Entrada::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
		
		//-- Obtener el objeto que obtendr la firma
		ArangiClientSignatureFactory signatureFactory = getClientSignatureFactory (signatureType);
		logger.debug ("[ArangiClient.counterSign]::Se ha obtenido el objeto que realizar la contrafirma: " + signatureFactory);
		
		//-- Elegir la firma que se quiere contrafirmar dentro del objeto firma
		SignatureSelector signatureSelector;
		try {
			signatureSelector = new SignatureSelector (i18n, signatureFactory.getSignatureObject(signatureFile));
		} catch (Exception e) {
			logger.error("[ArangiClient.counterSign]::No se ha podido obtener el objeto firma del tipo '" + signatureType + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", signatureType), e);
		}
		Certificate signatureToCounterSignCertificate = signatureSelector.getCertificateSignature ();
		
		//-- Elegir el certificado para realizar la contrafirma
		CertificateSelector certificateSelector = new CertificateSelector(i18n, filter, mostrarSeleccionSiSoloUno);
		DeviceAndAlias signatureCertificate = certificateSelector.getDeviceAndAlias();
		
		DeviceManager manager = null;
		try {
			//-- Obtener el manager
			manager = getManager (signatureCertificate);
			logger.debug("[ArangiClient.counterSign]::Obtenido el manager");
			
			//-- Contrafirmar
			signatureFactory.setSignatureOptions(this.options);
			signatureFactory.setUrlTSA(this.urlTSA);
			return signatureFactory.counterSign(signatureFile, manager, signatureCertificate.getAlias(), signatureToCounterSignCertificate);
			
		} finally {
			//-- Siempre cerrar el manager
			if (manager != null) {
				manager.close();
			}
		}
	}

	/**
	 * Valida una firma attached: CMS, PKCS#7, PDF, PADES-LTV, XAdES-BES attached,
	 * XAdES-T attached y XAdES-X-L attached.
	 * 
	 * @param signature Fichero con la firma
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Objetos de validacin (uno por cada firmador)
	 * @throws SignatureNotFoundException No existe el fichero pasado como parmetro
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 * @throws NoDetachedDocumentToValidateException La firma es detached y el documento no 
	 * 	existe o no puede ser ledo
	 */
	public ValidationObject validateSignature (File signature, String signatureType) throws SignatureNotFoundException, SignatureTypeUnknownException, SignatureFormatException, NoDetachedDocumentToValidateException {
		logger.info("[ArangiClient.validateSignature]::Entrada::" + Arrays.asList(new Object[] { signature, signatureType }));

		//-- Obtener el array de bytes
		if (signature == null || !signature.exists()) {
			logger.error("[ArangiClient.validateSignature]::No existe el fichero de firma: " + signature);
			throw new SignatureNotFoundException(i18n.getString("error.no.firma.validar"));
		}
		
		//-- Obtener la validacin
		ArangiClientValidationFactory clientValidation = getClientValidationFactory (signatureType);
		try {
			logger.debug ("[ArangiClient.validateSignature]::Se va a realizar la validacin la validacin: " + signatureType);
			return clientValidation.validate (signature);
		} catch (Exception e) {
			logger.error("[ArangiClient.validateSignature]::Error durante la validacin de la firma", e);
			throw new SignatureFormatException(i18n.getString("error.validando.firma"), e);
		}
		
	}
	
	/**
	 * Valida una firma attached: CMS, PKCS#7, PDF, PADES-LTV, XAdES-BES attached,
	 * XAdES-T attached y XAdES-X-L attached.
	 * 
	 * @param signature Fichero con la firma
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Objetos de validacin (uno por cada firmador)
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 */
	public ValidationObject validateSignature (byte[] signature, String signatureType) throws SignatureTypeUnknownException, SignatureFormatException {
		logger.info("[ArangiClient.validateSignature]::Entrada::" + Arrays.asList(new Object[] { signature, signatureType }));
		
		//-- Obtener la validacin
		ArangiClientValidationFactory clientValidation = getClientValidationFactory (signatureType);
		try {
			logger.debug ("[ArangiClient.validateSignature]::Se va a realizar la validacin la validacin: " + signatureType);
			return clientValidation.validate (signature);
		} catch (Exception e) {
			logger.error("[ArangiClient.validateSignature]::Error durante la validacin de la firma", e);
			throw new SignatureFormatException(i18n.getString("error.validando.firma"), e);
		}
		
	}
	
	/**
	 * Valida una firma attached: CMS, PKCS#7, PDF, PADES-LTV, XAdES-BES attached,
	 * XAdES-T attached y XAdES-X-L attached.
	 * 
	 * @param isSignature Stream al fichero con la firma
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Objetos de validacin (uno por cada firmador)
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 */
	public ValidationObject validateSignature (InputStream isSignature, String signatureType) throws SignatureTypeUnknownException, SignatureFormatException {
		logger.info("[ArangiClient.validateSignature]::Entrada::" + Arrays.asList(new Object[] { isSignature, signatureType }));
		
		//-- Obtener la validacin
		ArangiClientValidationFactory clientValidation = getClientValidationFactory (signatureType);
		try {
			logger.debug ("[ArangiClient.validateSignature]::Se va a realizar la validacin la validacin: " + signatureType);
			return clientValidation.validate (isSignature);
		} catch (Exception e) {
			logger.error("[ArangiClient.validateSignature]::Error durante la validacin de la firma", e);
			throw new SignatureFormatException(i18n.getString("error.validando.firma"), e);
		}
		
	}
	
	/**
	 * Valida una firma. Para el caso de las firmas attached no es necesario pasar el
	 * fichero detached.
	 * 
	 * @param signature Fichero con la firma
	 * @param detachedFile Fichero firmado (slo obligatorio para firmas detached)
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Objetos de validacin (uno por cada firmador)
	 * @throws SignatureNotFoundException No existe el fichero pasado como parmetro
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 * @throws NoDetachedDocumentToValidateException La firma es detached y el documento no 
	 * 	existe o no puede ser ledo
	 */
	public ValidationObject validateSignature (File signature, File detachedFile, String signatureType) throws SignatureNotFoundException, SignatureTypeUnknownException, SignatureFormatException, NoDetachedDocumentToValidateException {
		
		if (detachedFile == null) {
			return validateSignature(signature, signatureType);
		}
		
		try {
			return validateSignature(signature, Util.loadFile(detachedFile), signatureType);
		} catch (IOException e) {
			logger.error ("[ArangiClient.validateSignature]::El fichero detached no puede ser ledo", e);
			return validateSignature(signature, signatureType);
		}
	}
	
	/**
	 * Valida una firma. Para el caso de las firmas attached no es necesario pasar el
	 * contenido del fichero detached.
	 * 
	 * @param signature Fichero con la firma
	 * @param detachedContent Fichero firmado (slo para firmas detached)
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Objetos de validacin (uno por cada firmador)
	 * @throws SignatureNotFoundException No existe el fichero pasado como parmetro
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 */
	public ValidationObject validateSignature (File signature, byte[] detachedContent, String signatureType) throws SignatureNotFoundException, SignatureTypeUnknownException, SignatureFormatException {
		logger.info("[ArangiClient.validateSignature]::Entrada::" + Arrays.asList(new Object[] { signature, detachedContent, signatureType }));

		//-- Obtener el array de bytes
		if (signature == null || !signature.exists()) {
			logger.error("[ArangiClient.validateSignature]::No existe el fichero de firma: " + signature);
			throw new SignatureNotFoundException(i18n.getString("error.no.firma.validar"));
		}
		
		//-- Obtener el documento detached
		ByteArrayDocument document = new ByteArrayDocument(detachedContent);
		
		//-- Obtener la validacin
		ArangiClientValidationFactory clientValidation = getClientValidationFactory (signatureType);
		try {
			logger.debug ("[ArangiClient.validateSignature]::Se va a realizar la validacin la validacin: " + signatureType);
			return clientValidation.validate (signature, document);
		} catch (Exception e) {
			logger.error("[ArangiClient.validateSignature]::Error durante la validacin de la firma", e);
			throw new SignatureFormatException(i18n.getString("error.validando.firma"), e);
		}
		
	}
	
	/**
	 * Valida una firma. Para el caso de las firmas attached no es necesario pasar el
	 * fichero detached.
	 * 
	 * @param signature Contenido de la firma
	 * @param detachedFile Fichero firmado (slo obligatorio para firmas detached)
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Objetos de validacin (uno por cada firmador)
	 * @throws SignatureNotFoundException No existe el fichero pasado como parmetro
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 * @throws NoDetachedDocumentToValidateException La firma es detached y el documento no 
	 * 	existe o no puede ser ledo
	 */
	public ValidationObject validateSignature (byte[] signature, File detachedFile, String signatureType) throws SignatureNotFoundException, SignatureTypeUnknownException, SignatureFormatException, NoDetachedDocumentToValidateException {
		
		if (detachedFile == null) {
			return validateSignature(signature, signatureType);
		}
		
		try {
			return validateSignature(signature, Util.loadFile(detachedFile), signatureType);
		} catch (IOException e) {
			logger.error ("[ArangiClient.validateSignature]::El fichero detached no puede ser ledo", e);
			return validateSignature(signature, signatureType);
		}
	}
	
	/**
	 * Valida una firma. Para el caso de las firmas attached no es necesario pasar el
	 * contenido del fichero detached.
	 * 
	 * @param signature Contenido de la firma
	 * @param detachedContent Fichero firmado (slo para firmas detached)
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return Objetos de validacin (uno por cada firmador)
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 */
	public ValidationObject validateSignature (byte[] signature, byte[] detachedContent, String signatureType) throws SignatureTypeUnknownException, SignatureFormatException {
		logger.info("[ArangiClient.validateSignature]::Entrada::" + Arrays.asList(new Object[] { signature, detachedContent, signatureType }));

		//-- Obtener el documento detached
		ByteArrayDocument document = new ByteArrayDocument(detachedContent);
		
		//-- Obtener la validacin
		ArangiClientValidationFactory clientValidation = getClientValidationFactory (signatureType);
		try {
			logger.debug ("[ArangiClient.validateSignature]::Se va a realizar la validacin la validacin: " + signatureType);
			return clientValidation.validate (signature, document);
		} catch (Exception e) {
			logger.error("[ArangiClient.validateSignature]::Error durante la validacin de la firma", e);
			throw new SignatureFormatException(i18n.getString("error.validando.firma"), e);
		}
	}
	
	/**
	 * Valida una firma. Para el caso de las firmas attached no es necesario pasar el
	 * contenido del fichero detached.
	 * 
	 * @param isSignature Stream al fichero con la firma
	 * @param detachedContent Stream al fichero firmado (slo para firmas detached)
	 * @param signatureType Tipo de firma segn las constantes definidas en la clase 
	 * 	{@link Constants Constants}
	 * @return XML resultado de la validacin
	 * @throws SignatureTypeUnknownException  No se ha encontrado una clase capaz de realizar
	 * 	la firma demandada
	 * @throws SignatureFormatException La firma no ha podido ser cargada
	 */
	public ValidationObject validateSignature (InputStream isSignature, InputStream detachedContent, String signatureType) throws SignatureTypeUnknownException, SignatureFormatException {
		logger.info("[ArangiClient.validateSignature]::Entrada::" + Arrays.asList(new Object[] { isSignature, detachedContent, signatureType }));

		//-- Obtener el documento detached
		InputStreamDocument document = new InputStreamDocument(detachedContent);
		
		//-- Obtener la validacin
		ArangiClientValidationFactory clientValidation = getClientValidationFactory (signatureType);
		try {
			logger.debug ("[ArangiClient.validateSignature]::Se va a realizar la validacin la validacin: " + signatureType);
			return clientValidation.validate (isSignature, document);
		} catch (Exception e) {
			logger.error("[ArangiClient.validateSignature]::Error durante la validacin de la firma", e);
			throw new SignatureFormatException(i18n.getString("error.validando.firma"), e);
		}
	}
	
	/**
	 * Valida un certificado y devuelve un XML con la respuesta
	 * 
	 * @param isCertificate Stream de lectura del certificado
	 * @return XML con la respuesta
	 * @throws InvalidCertificateException El parmetro no contiene un certificado digital
	 * 	vlido
	 * @throws NoArangiCertificateException El certificado no es uno de los tratados por
	 * 	Arang
	 * @throws UnsupportedEncodingException La codificacin UTF-8 no est soportada
	 */
	public String validateCertificate (InputStream isCertificate) throws MalformedCertificateException, NoArangiCertificateException, UnsupportedEncodingException {
		logger.info("[ArangiClient.validateCertificate]::Entrada::" + Arrays.asList(new Object[] { isCertificate }));
		
		X509Certificate userCertificate;
		try {
			userCertificate = Util.getCertificate(isCertificate);
		} catch (NormalizeCertificateException e) {
			logger.warn("[ArangiClient.validateCertificate]::El certificado no puede ser normalizado con las libreras BouncyCastle", e);
        	throw new MalformedCertificateException(i18n.getString("error.normalizando.certificado"), e);
		}
		if (userCertificate == null) {
			logger.warn("[ArangiClient.validateCertificate]::El contenido pasado no es un certificado digital");
        	throw new MalformedCertificateException(i18n.getString("error.no.es.certificado"));
		}
		
		//-- Se ha podido cargar el certificado
		ValidateCertificate validateCertificate = null;
		try {
			validateCertificate = CertificateFactory.getInstance(new Certificate(userCertificate));
		} catch (NormalizeCertificateException e) {
			logger.warn("[ArangiClient.validateCertificate]::El certificado no puede ser normalizado con las libreras BouncyCastle", e);
        	throw new MalformedCertificateException(i18n.getString("error.normalizando.certificado"), e);
		}
		
		if (validateCertificate == null) {
			logger.warn("[ArangiClient.validateCertificate]::El certificado no es de los tratados por Arang::" + userCertificate);
			throw new NoArangiCertificateException(i18n.getString("error.certificado.no.arangi", userCertificate.getIssuerDN()));
		}
		
		//-- El certificado es correcto -> validar
		StringBuffer sb = new StringBuffer("");
		sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		sb.append("<certificate_result>");
		ValidatingTrace trace = new ValidatingTrace();
		int result = validateCertificate.validate(trace);
		sb.append ("<valid>" + (result == ValidationResult.RESULT_VALID) + "</valid>");
		sb.append ("<result>" + result + "</result>");
		sb.append ("<result_message>" + CertificateValidator.getString(result) + "</result_message>");
		sb.append ("<date>" + DATE_FORMAT.format(new Date()) + "</date>");
//		request.setAttribute("traza", trace.toHTML());
		
		//-- Informacin del certificado
		sb.append("<certificate>");
		sb.append(getCertificateInformation(validateCertificate));
		sb.append("</certificate>");
		
		//-- Informacin del propietario del certificado
		sb.append("<subject_information>");
		SignerInformation information = new SignerInformation(validateCertificate);
		sb.append(getSubjectCertificateInformation(information));
		sb.append("</subject_information>");
		
		sb.append("</certificate_result>");
		logger.debug("[ArangiClient.validateCertificate]::Respuesta::" + sb.toString());
		try {
			return new String (sb.toString().getBytes("UTF-8"), "UTF-8"); 
		} catch (IOException e) {
			logger.warn("[ArangiClient.validateCertificate]::Error codificando a UTF-8", e);
			throw new UnsupportedEncodingException(i18n.getString("error.codificacion.texto", "UTF-8"));
		}

	}
	
	/**
	 * A partir de los resultados de una validacin obtiene un xml 
	 * 
	 * @param validationObjects Resultado de la validacin
	 * @return XML con los resultados de la validacin
	 * @throws UnsupportedEncodingException La codificacin UTF-8 no est soportada
	 */
	public String getSignatureValidationResultXML(ValidationObject validationObject) throws UnsupportedEncodingException {
		StringBuffer sb = new StringBuffer ();
		sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		sb.append("<validation_result>");
		sb.append ("<signature_type>" + validationObject.getSignatureType() + "</signature_type>");
		sb.append ("<date>" + DATE_FORMAT.format(new Date()) + "</date>");
		sb.append("<signatures>");
		for (int i = 0; i < validationObject.getSignatureValidationResults().length; i++) {
			sb.append ("<signature>");
			sb.append (getTextValidationForSignature (validationObject.getSignatureValidationResults()[i]));
			sb.append ("</signature>");
		}
		
		sb.append("</signatures></validation_result>");
		try {
			return new String (sb.toString().getBytes("UTF-8"), "UTF-8"); 
		} catch (IOException e) {
			logger.warn("[ArangiClient.validateCertificate]::Error codificando a UTF-8", e);
			throw new UnsupportedEncodingException(i18n.getString("error.codificacion.texto", "UTF-8"));
		}
	}

	/**
	 * Obtiene el objeto que construir la firma
	 * 
	 * @param signatureType Tipo de firma
	 * @return Objeto que tratar la firma
	 * @throws SignatureTypeUnknownException No hay ningn objeto para tratar ese tipo de firma
	 */
	public ArangiClientSignatureFactory getClientSignatureFactory(String signatureType) throws SignatureTypeUnknownException {
		logger.debug("[ArangiClient.getClientSignatureFactory]::Entrada::" + signatureType);
		
		//-- Obtener el nombre de la clase
		String className = "";
		StringTokenizer st = new StringTokenizer(signatureType, "-");
		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			className += token.substring(0,1).toUpperCase() + token.substring(1);
		}
		className = SIGNATURE_PACKAGE + "." + className + "SignatureFactory";
		logger.debug("[ArangiClient.getClientSignatureFactory]::Obteniendo objeto: " + className);
		
		//-- Obtener el objeto
		try {
			return (ArangiClientSignatureFactory) Class.forName(className).getConstructor(new Class[] { Boolean.class, I18n.class }).newInstance( signatureType.indexOf("attached")>-1, i18n );
		} catch (Exception e) {
			// No se debe dar a no ser que est mal programada la clase
			logger.error("[ArangiClient.getClientSignatureFactory]::Error obteniendo el objeto que firmar: " + className, e);
			throw new SignatureTypeUnknownException("Error obteniendo el objeto que firmar: " + className, e);
		}		
	}

	/**
	 * Obtiene el objeto que validar un tipo de firma
	 * 
	 * @param signatureType Tipo de firma
	 * @return Objeto que validar un tipo de firmas
	 * @throws SignatureTypeUnknownException No hay ningn objeto para tratar ese tipo de firma
	 */
	public ArangiClientValidationFactory getClientValidationFactory(String signatureType) throws SignatureTypeUnknownException {
		logger.debug("[ArangiClient.getClientValidationFactory]::Entrada::" + signatureType);
		
		//-- Obtener el nombre de la clase
		String className = "";
		StringTokenizer st = new StringTokenizer(signatureType, "-");
		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			className += token.substring(0,1).toUpperCase() + token.substring(1);
		}
		className = VALIDATION_PACKAGE + "." + className + "ValidationFactory";
		logger.debug("[ArangiClient.getClientValidationFactory]::Obteniendo objeto: " + className);
		
		//-- Obtener el objeto
		try {
			return (ArangiClientValidationFactory) Class.forName(className).getConstructor(new Class[] { I18n.class }).newInstance( i18n );
		} catch (Exception e) {
			// No se debe dar a no ser que est mal programada la clase
			logger.error("[ArangiClient.getClientValidationFactory]::Error obteniendo el objeto que validar: " + className, e);
			throw new SignatureTypeUnknownException("Error obteniendo el objeto que validar: " + className, e);
		}		
	}

	
	//-- Mtodos privados

	/*
	 * Obtiene un manager a partir de la informacin pasada como parmetro
	 */
	private DeviceManager getManager(DeviceAndAlias signCertificate) throws OpeningDeviceException {
		logger.info ("[ArangiClient.getManager]::Entrada");
		
		//-- Keystore MOZILLA
		if (signCertificate.getKeystoreType().equals(CertificateSelector.MOZILLA)) {
			logger.debug ("[ArangiClient.getManager]::El keystore del certificado es MOZILLA");
			try {
				return new MozillaKeyStoreManager();
			} catch (MozillaNotFoundException e) {
				logger.info("[ArangiClient.getManager]::Mozilla no est instalado en el equipo");
				throw new OpeningDeviceException("Mozilla no est instalado en el equipo");
			} catch (MozillaKeyStoreException e) {
				logger.error("[ArangiClient.getManager]::Error abriendo el almacn de claves de Mozilla", e);
				throw new OpeningDeviceException("Error abriendo el almacn de claves de Mozilla", e);
			}
		}

		//-- Keystore IEXPLORER
		if (signCertificate.getKeystoreType().equals(CertificateSelector.IEXPLORER)) {
			logger.debug ("[ArangiClient.getManager]::El keystore del certificado es IEXPLORER");
			try {
				return new PersonalWindowsKeyStoreManager();
			} catch (OpeningDeviceException e) {
				logger.error("[ArangiClient.getManager]::Error abriendo el almacn de claves de IExplorer", e);
				throw new OpeningDeviceException("Error abriendo el almacn de claves de IExplorer", e);
			} catch (WindowsNotFoundException e) {
				logger.error("[ArangiClient.getManager]::La ejecucin no se encuentra en un sistema operativo Windows", e);
				throw new OpeningDeviceException("La ejecucin no se encuentra en un sistema operativo Windows", e);
			}
		}

		//-- Keystore Apple Safari
		if (signCertificate.getKeystoreType().equals(CertificateSelector.APPLE)) {
			logger.debug ("[ArangiClient.getManager]::El keystore del certificado es APPLE");
			try {
				return new AppleSafariManager();
			} catch (AppleSafariKeyStoreException e) {
				logger.error("[ArangiClient.getManager]::Error abriendo el almacn de claves de Apple Safari", e);
				throw new OpeningDeviceException("Error abriendo el almacn de claves de Apple Safari", e);
			} catch (AppleSafariNotFoundException e) {
				logger.error("[ArangiClient.getManager]::La ejecucin no se encuentra en un sistema operativo Mac OS X", e);
				throw new OpeningDeviceException("La ejecucin no se encuentra en un sistema operativo Mac OS X", e);
			}
		}

		//-- Keystore PKCS#11
		logger.debug ("[ArangiClient.getManager]::El keystore del certificado es PKCS#11");
		
		//-- Preguntar por el pin
		String pin = null;
		while (true) {
			pin = (String)JOptionPane.showInputDialog(null, i18n.getString("titulo.escribir.pin.pkcs11") + ": " + signCertificate.getPkcs11Device().getManufacturerId(), "PIN");
			if (pin == null || pin.length() == 0) {
				 int response = JOptionPane.showConfirmDialog(null, "???", i18n.getString("pregunta.no.escribir.pin.pkcs11"), JOptionPane.YES_NO_OPTION);
				 if (response == JOptionPane.OK_OPTION) {
					 break;
				 }
			} else {
				try {
					return new Pkcs11Manager(signCertificate.getPkcs11Device(), pin);
				} catch (ModuleNotFoundException e) {
					//-- No se dar porque ya se ha detectado la tarjeta con el mdulo
					logger.error("[CertificateSelector.getDeviceAndAlias]::No se detecta el mdulo que debe abrir la tarjeta: " + signCertificate.getPkcs11Device().getManufacturerId(), e);
					JOptionPane.showMessageDialog(null, i18n.getString("error.modulo.desconocido"));
					break;
				} catch (IncorrectPINException e) {
					logger.error("[CertificateSelector.getDeviceAndAlias]::El PIN es incorrecto");
					JOptionPane.showMessageDialog(null, i18n.getString("error.pin.incorrecto"));
				} catch (LockedPINException e) {
					logger.error("[CertificateSelector.getDeviceAndAlias]::La tarjeta est bloqueada");
					JOptionPane.showMessageDialog(null, i18n.getString("error.tarjeta.bloqueada"));
					break;
				} catch (OpeningDeviceException e) {
					logger.error("[CertificateSelector.getDeviceAndAlias]::Error abriendo la tarjeta", e);
					JOptionPane.showMessageDialog(null, i18n.getString("error.abriendo.tarjeta"));
				} catch (DeviceNotFoundException e) {
					logger.error("[CertificateSelector.getDeviceAndAlias]::Error abriendo el almacn de claves de IExplorer", e);
					JOptionPane.showMessageDialog(null, i18n.getString("error.tarjeta.extraida"));
				} catch (IAIKDLLNotFoundException e) {
					logger.error("[CertificateSelector.getDeviceAndAlias]::No se encuentra la DLL de IAIK", e);
					JOptionPane.showMessageDialog(null, i18n.getString("error.abriendo.tarjeta"));
					break;
				} catch (InitializeProviderException e) {
					logger.error("[CertificateSelector.getDeviceAndAlias]::No se puede inicializar el proveedor criptogrfico", e);
					JOptionPane.showMessageDialog(null, i18n.getString("error.abriendo.tarjeta"));
					break;
				}
			}
		}
		
		//-- No se ha podido abrir el dispositivo
		logger.error("[CertificateSelector.getDeviceAndAlias]::No se ha podido abrir la tarjeta");
		throw new OpeningDeviceException("No se ha podido abrir la tarjeta");
	}

	/*
	 * Obtiene un xml para una firma
	 */
	private StringBuffer getTextValidationForSignature(ValidationObject.SignatureValidationResult signatureValidationResult) {
		
		StringBuffer sb = new StringBuffer ("");
		sb.append ("<valid>" + signatureValidationResult.isValid() + "</valid>");
		sb.append ("<result>" + signatureValidationResult.getResult() + "</result>");
		sb.append ("<result_message>" + (signatureValidationResult.isValid()?"":signatureValidationResult.getInvalidMessage()) + "</result_message>");
		
		// Fecha
		if (signatureValidationResult.getDate() != null) {
			sb.append ("<date>" + DATE_FORMAT.format(signatureValidationResult.getDate()) + "</date>");
		}
		
		// Certificado de firma
		ValidateCertificate certificate = signatureValidationResult.getCertificate();
		if (certificate != null) {
			sb.append("<certificate>");
			sb.append(getCertificateInformation(certificate));
			sb.append("</certificate>");
		}
		
		// Informacin del certificado de firma
		SignerInformation information = signatureValidationResult.getSignerInformation();
		if (information != null) {
			sb.append("<signer_information>");
			sb.append(getSubjectCertificateInformation(information));
			sb.append("</signer_information>");
		}
		
		// sello de tiempos (si existe)
		if (signatureValidationResult.getTimeStamp() != null) {
			sb.append("<timestamp>");
			Certificate tsCertificate = null;
			try {
				tsCertificate = signatureValidationResult.getTimeStamp().getSignatureCertificate();
			} catch (MalformedTimeStampException e) {
				logger.info("[ArangiClientValidationApplet.getTextValidation]::El sello de tiempos est mal formado", e);
			}
			if (tsCertificate != null) {
				sb.append("<timestamp_certificate>");
				sb.append(getCertificateInformation(tsCertificate));
				sb.append("</timestamp_certificate>");
			}
			sb.append ("<date>" + DATE_FORMAT.format(signatureValidationResult.getTimeStamp().getTime()) + "</date>");
			sb.append("</timestamp>");
		}
		
		// ocsps (si existen)
		if (signatureValidationResult.getOcspResponses() != null && signatureValidationResult.getOcspResponses().length > 0) {
			sb.append("<ocsps>");
			OCSPResponse[] ocspResponses = signatureValidationResult.getOcspResponses();
			for (int j = 0; j < ocspResponses.length; j++) {
				sb.append("<ocsp>");
				Certificate ocspCertificate = ocspResponses[j].getSignatureCertificate();
				if (ocspCertificate != null) {
					sb.append("<ocsp_certificate>");
					sb.append(getCertificateInformation(ocspCertificate));
					sb.append("</ocsp_certificate>");
				}
				sb.append("</ocsp>");
			}
			sb.append("</ocsps>");
		}
		
		// contrafirmas
		if (signatureValidationResult.getCounterSignatures().length > 0) {
			sb.append("<countersignatures>");
			for (int j = 0; j < signatureValidationResult.getCounterSignatures().length; j++) {
				sb.append("<countersignature>");
				sb.append(getTextValidationForSignature(signatureValidationResult.getCounterSignatures()[j]));
				sb.append("</countersignature>");
			}
			sb.append("</countersignatures>");
		}
		
		
		return sb;
	}

	/*
	 * Obtiene la informacin del XML correspondiente a un certificado
	 */
	private static String getCertificateInformation (Certificate certificate) {
		StringBuffer sb = new StringBuffer("");
		sb.append("<common_name>");
		sb.append (certificate.getCommonName());
		sb.append("</common_name>");
		sb.append("<validity>");
		sb.append("<not_before>");
		sb.append (DATE_FORMAT.format(certificate.getValidityPeriodBeginning()));
		sb.append("</not_before>");
		sb.append("<not_after>");
		sb.append (DATE_FORMAT.format(certificate.getValidityPeriodEnd()));
		sb.append("</not_after>");
		sb.append("</validity>");
		sb.append("<issuer>");
		sb.append (certificate.getIssuerCommonName());
		sb.append("</issuer>");

		return sb.toString();
	}

	/*
	 * Obtiene la informacin del propietario del certificado
	 */
	private String getSubjectCertificateInformation(SignerInformation information) {
		StringBuffer sb = new StringBuffer("");
		if(information.getTipoCertificado() != null) {
			sb.append("<type>" + i18n.getString(information.getTipoCertificado()) + "</type>");
		}
		if(information.getNif() != null) {
			sb.append("<nif>" + information.getNif() + "</nif>");
		}
		if(information.getCif() != null) {
			sb.append("<cif>" + information.getCif() + "</cif>");
		}
		if(information.getName() != null) {
			sb.append("<name>" + information.getName() + "</name>");
		}
		if(information.getEmail() != null) {
			sb.append("<email>" + information.getEmail() + "</email>");
		}
		if(information.getPosition() != null) {
			sb.append("<position>" + information.getPosition() + "</position>");
		}
		if(information.getEntity() != null) {
			sb.append("<entity>" + information.getEntity() + "</entity>");
		}
		if(information.getOrganizationalUnit() != null) {
			sb.append("<organizational_unit>" + information.getOrganizationalUnit() + "</organizational_unit>");
		}
		if(information.getNrpnip() != null) {
			sb.append("<nrpnip>" + information.getNrpnip() + "</nrpnip>");
		}
		if(information.getAgentNIF() != null) {
			sb.append("<agent_nif>" + information.getAgentNIF() + "</agent_nif>");
		}
		if(information.getAgentName() != null) {
			sb.append("<agent_name>" + information.getAgentName() + "</agent_name>");
		}
		return sb.toString();
	}


}
