package es.accv.arangi.client;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.io.File;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import javax.swing.JFileChooser;

import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.client.filter.ArangiClientFilter;
import es.accv.arangi.client.util.I18n;
import es.accv.arangi.timestamp.TimeStamp;

/**
 * Applet de firma del cliente de Arang
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class ArangiClientSignatureApplet extends ArangiClientApplet {

	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 1L;
	
	/*
	 * Cliente de Arang
	 */
	private ArangiClient client;
	
	/*
	 * Tipo de firma
	 */
	private String signatureType;
	
	/*
	 * Filtro para seleccionar certificados
	 */
	private ArangiClientFilter filter;
	
	/*
	 * Fichero a firmar
	 */
	private File fileToSign;
	
	/*
	 * Contenido a firmar
	 */
	private byte[] contentToSign;
	
	/*
	 * Fichero con la firma (para contrafirmar)
	 */
	private File signatureFile;
	
	/*
	 * Contenido de la firma (para contrafirmar)
	 */
	private byte[] signatureContent;
	
	/*
	 * URL a la TSA
	 */
	private String urlTSA = TimeStamp.URL_ACCV_TSA_SERVER;
	
	/*
	 * Opciones extra para aadir a las firmas
	 */
	private Map<String, Object> options = new HashMap<String, Object>();
	
	/*
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero elegido.
	 */
	private String reference;
	
	/*
	 * Resultado en base 64 de alguno de los mtodos de firma. Se ir pasando trozo a trozo
	 * a javascript para evitar problemas con ficheros grandes
	 */
	private SignatureBase64Stream resultB64;

	
	/**
	 * Constructor: carga valores por defecto
	 */
	public ArangiClientSignatureApplet () {
		// La firma por defecto es XAdES-BES attached
		this.signatureType = SignatureConstants.SIGNATURE_XADES_BES_ATTACHED;
		
		this.filter = new ArangiClientFilter();
	}
	
	@Override
	public void init() {
		super.init(); 
		
		//-- Obtener los parmetros de pas y lengua
		String language = getParameter("language");
		String country = getParameter("country");
		
		//-- Si no existen, por defecto espaol
		if (language == null) {
			language = "es";
		}
		if (country == null) {
			country = "ES";
		}
		
		//-- Cargar el cliente de Arang y la internacionalizacin
		client = new ArangiClient(language, country);		
		this.i18n = new I18n (new Locale(language, country)); 

		//-- Centrar
		centerComponent(this);
	}
	
	//-- Mtodos pblicos
	
	/**
	 * Carga el tipo de firma a utilizar
	 */
	public void setSignatureType (String signatureType) {
		this.signatureType = signatureType;
	}

	/**
	 * Carga el fichero a firmar
	 * 
	 * @param filePath Path al fichero a firmar
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero elegido.
	 */
	public void setFileToSign (String filePath, String reference) {
		this.fileToSign = new File (filePath);
		this.contentToSign = null;
		this.reference = reference;
	}
	
	/**
	 * Carga el contenido a firmar
	 * 
	 * @param content Contenido a firmar
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero elegido.
	 */
	public void setContentToSign (String content, String reference) {
		this.contentToSign = content.getBytes();
		this.fileToSign = null;
		this.reference = reference;
	}
	
	/**
	 * Carga el contenido a firmar (el contenido se pasa en base64)
	 * 
	 * @param content Contenido a firmar (se pasa en base64)
	 * @param reference Referencia incluida en las firmas XAdES detached o motivo de firma
	 * 	para los PDFs. Si se deja a null en una firma XAdES detached la referencia de la
	 * 	firma ser el path del fichero elegido.
	 */
	public void setContentToSignBase64 (String contentB64, String reference) {
		this.contentToSign = Util.decodeBase64(contentB64);
		this.fileToSign = null;
		this.reference = reference;
	}
	
	/**
	 * Carga el fichero con la firma
	 * 
	 * @param filePath Path al fichero con la firma
	 */
	public void setSignatureFile (String filePath) {
		this.signatureFile = new File (filePath);
		this.signatureContent = null;
	}
	
	/**
	 * Carga el contenido de la firma
	 * 
	 * @param content Contenido de la firma
	 */
	public void setSignatureContent (String content) {
		this.signatureContent = content.getBytes();
		this.signatureFile = null;
	}
	
	/**
	 * Carga el contenido de la firma (el contenido se pasa en base64)
	 * 
	 * @param content Contenido de la firma (se pasa en base64)
	 */
	public void setSignatureContentBase64 (String contentB64) {
		this.signatureContent = Util.decodeBase64(contentB64);
		this.signatureFile = null;
	}
	
	/**
	 * Aade una regla sobre el DN del asunto para el filtro de certificados
	 * 
	 * @param rule Regla a aadir
	 */
	public void addSubjectDNFilterRule (String rule) {
		filter.addSubjectDNFilterRule(rule);
	}

	/**
	 * Aade una regla sobre el DN del emisor para el filtro de certificados
	 * 
	 * @param rule Regla a aadir
	 */
	public void addIssuerDNFilterRule (String rule) {
		filter.addIssuerDNFilterRule(rule);
	}

	/**
	 * Aade una regla sobre el nmero de serie para el filtro de certificados
	 * 
	 * @param rule Nmero de serie a aadir
	 */
	public void addSerialNumberFilterRule (String serialNumber) {
		filter.addSerialNumberFilterRule(serialNumber);
	}

	/**
	 * Le dice al filtro si se quiren todos los certificados independientemente
	 * de sus usos de clave. Si es false o no se llama a este mtodo slo se
	 * mostrarn los certificados de firma.
	 * 
	 * @param anyKeyUsage Todos los usos de clave
	 */
	public void allKeyUsagesFilter(boolean allKeyUsages) {
		filter.setAllKeyUsages(allKeyUsages);
	}
	
	/**
	 * Desactiva la bsqueda de certificados dentro del almacn de Mozilla
	 */
	public void disableMozillaKeystore () {
		filter.setMozillaKeystore(false);
	}
	
	/**
	 * Desactiva la bsqueda de certificados dentro del almacn de Internet Explorer (Windows)
	 */
	public void disableIExplorerKeystore () {
		filter.setIExplorerKeystore(false);
	}
	
	/**
	 * Desactiva la bsqueda de certificados dentro de los almacenes PKCS#11
	 */
	public void disablePkcs11Keystores() {
		filter.setPkcs11Keystore(false);
	}
	
	/**
	 * Usar slo el almacn de Mozilla
	 */
	public void useOnlyMozillaKeystore() {
		filter.setIExplorerKeystore(false);
		filter.setPkcs11Keystore(false);
	}
	
	/**
	 * Usar slo el almacn de Internet Explorer (Windows)
	 */
	public void useOnlyIExplorerKeystore() {
		filter.setMozillaKeystore(false);
		filter.setPkcs11Keystore(false);
	}
	
	/**
	 * Resetea la lista de filtros
	 */
	public void resetFilter () {
		filter = new ArangiClientFilter();
	}
	
	/**
	 * Carga la variable que indica si se muestra la seleccin si slo hay un
	 * elemento a seleccionar
	 */
	public void setMostrarSeleccionSiSoloUno(boolean mostrarSeleccionSiSoloUno) {
		client.setMostrarSeleccionSiSoloUno(mostrarSeleccionSiSoloUno);
	}
	
	/**
	 * Carga la variable que indica se queremos que la firma PDF lleve o no sello
	 * de tiempos
	 */
	public void setFirmaPDFSinSelloTiempos (boolean firmaPDFSinSelloTiempos) {
		addParametroExtraFirma (ArangiClient.OPCION_FIRMAR_SIN_SELLO, new Boolean(firmaPDFSinSelloTiempos));
	}
	
	/**
	 * Aade un parmetro extra que se aadir a la firma si el formato 
	 * lo acepta.
	 * 
	 * @param key Clave
	 * @param value Valor
	 */
	public void addParametroExtraFirma (String key, Object value) {
		options.put(key, value);
	}
	
	/**
	 * Cambia el valor de la TSA a utilizar dentro del cliente
	 */
	public void setUrlTSA (String urlTSA) {
		this.urlTSA = urlTSA;
	}
	
	/**
	 * Obtiene el texto del error producido
	 * 
	 * @return texto del error
	 */
	public String getError() {
		if (this.error == null) {
			return null;
		}
		return this.error.getMessage();
	}
	
	/**
	 * Devuelve la pila de la ltima excepcin en base64 
	 */
	public String getPilaError () {
		if (this.error == null) {
			return null;
		}
		return Util.encodeBase64(getFullStackTrace(this.error).toString().getBytes());
	}
	
	/**
	 * Realiza la firma
	 * 
	 * @return Firma en base64
	 */
	public String sign () {
		logger.info("[ArangiClientSignatureApplet.sign]::Entrada");
		this.error = null;
		
		//-- Pasar las opciones extra y la URL de la TSA
		client.setOpciones(this.options);
		client.setUrlTSA(this.urlTSA);
				
		try {
			if (this.fileToSign != null) {
				logger.info("[ArangiClientSignatureApplet.sign]::Firmando un fichero::" + Arrays.asList(new Object[] { fileToSign, filter, reference, signatureType }));
				return client.signFile(fileToSign, filter, reference, signatureType);
			}
			if (this.contentToSign != null) {
				logger.info("[ArangiClientSignatureApplet.sign]::Firmando contenido::" + Arrays.asList(new Object[] { contentToSign, filter, reference, signatureType }));
				return client.signContent(contentToSign, filter, reference, signatureType);
			}
			
			//-- Se ha llegado hasta aqu (no se ha definido nada para firmar
			logger.error("[ArangiClientSignatureApplet.sign]::No se ha definido nada que firmar");
			throw new NoDocumentToSignException (i18n.getString("error.no.documento.firmar"));
		} catch (Throwable e) {
			logger.error ("[ArangiClientSignatureApplet.sign]::ERROR firmando", e);
			this.error = e;
			return null;
		}
		
	}
	
	
	/**
	 * Realiza la firma. Luego hay que llamar al mtodo getResultPiece para ir obteniendo la firma
	 */
	public void signReturnStream () {
		logger.info("[ArangiClientSignatureApplet.signReturnStream]::Entrada");
		this.error = null;
		
		//-- Pasar las opciones extra y la URL de la TSA
		client.setOpciones(this.options);
		client.setUrlTSA(this.urlTSA);
				
		try {
			if (this.fileToSign != null) {
				logger.info("[ArangiClientSignatureApplet.signReturnStream]::Firmando un fichero::" + Arrays.asList(new Object[] { fileToSign, filter, reference, signatureType }));
				resultB64 = new SignatureBase64Stream(client.signFile(fileToSign, filter, reference, signatureType));
				return;
			}
			if (this.contentToSign != null) {
				logger.info("[ArangiClientSignatureApplet.signReturnStream]::Firmando contenido::" + Arrays.asList(new Object[] { contentToSign, filter, reference, signatureType }));
				resultB64 = new SignatureBase64Stream(client.signContent(contentToSign, filter, reference, signatureType));
				return;
			}
			
			//-- Se ha llegado hasta aqu (no se ha definido nada para firmar
			logger.error("[ArangiClientSignatureApplet.signReturnStream]::No se ha definido nada que firmar");
			throw new NoDocumentToSignException (i18n.getString("error.no.documento.firmar"));
		} catch (Throwable e) {
			logger.error ("[ArangiClientSignatureApplet.signReturnStream]::ERROR firmando", e);
			this.error = e;
		}
		
	}
	
	/**
	 * Realiza la cofirma
	 * 
	 * @return Firma en base64
	 */
	public String coSign () {
		logger.info("[ArangiClientSignatureApplet.coSign]::Entrada");
		this.error = null;
		
		//-- Pasar las opciones extra
		client.setOpciones(this.options);
				
		try {
			if (this.signatureFile != null) {
				logger.info("[ArangiClientSignatureApplet.coSign]::Cofirmando firma en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
				if (this.fileToSign != null) {
					logger.info("[ArangiClientSignatureApplet.coSign]::Cofirmando documento en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					return client.coSign(signatureFile, fileToSign, filter, reference, signatureType);
				} else {
					logger.info("[ArangiClientSignatureApplet.coSign]::Cofirmando contenido de documento::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					return client.coSign(signatureFile, contentToSign, filter, reference, signatureType);
				}
			}
			if (this.signatureContent != null) {
				logger.info("[ArangiClientSignatureApplet.coSign]::Cofirmando contenido de una firma::" + Arrays.asList(new Object[] { signatureContent, filter, signatureType }));
				if (this.fileToSign != null) {
					logger.info("[ArangiClientSignatureApplet.coSign]::Cofirmando documento en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					return client.coSign(signatureContent, fileToSign, filter, reference, signatureType);
				} else {
					logger.info("[ArangiClientSignatureApplet.coSign]::Cofirmando contenido de documento::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					return client.coSign(signatureContent, contentToSign, filter, reference, signatureType);
				}
			}
			
			//-- Se ha llegado hasta aqu (no se ha definido nada para contrafirmar
			logger.error("[ArangiClientSignatureApplet.coSign]::No se ha definido nada que cofirmar");
			throw new NoDocumentToSignException (i18n.getString("error.no.firma.cofirmar"));
		} catch (Throwable e) {
			logger.error ("[ArangiClientSignatureApplet.coSign]::ERROR firmando", e);
			this.error = e;
			return null;
		}
		
	}
	
	/**
	 * Realiza la cofirma. Luego hay que llamar al mtodo getResultPiece para ir obteniendo la firma
	 */
	public void coSignReturnStream () {
		logger.info("[ArangiClientSignatureApplet.coSignReturnStream]::Entrada");
		this.error = null;
		
		//-- Pasar las opciones extra
		client.setOpciones(this.options);
				
		try {
			if (this.signatureFile != null) {
				logger.info("[ArangiClientSignatureApplet.coSignReturnStream]::Cofirmando firma en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
				if (this.fileToSign != null) {
					logger.info("[ArangiClientSignatureApplet.coSignReturnStream]::Cofirmando documento en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					resultB64 = new SignatureBase64Stream(client.coSign(signatureFile, fileToSign, filter, reference, signatureType));
				} else {
					logger.info("[ArangiClientSignatureApplet.coSignReturnStream]::Cofirmando contenido de documento::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					resultB64 = new SignatureBase64Stream(client.coSign(signatureFile, contentToSign, filter, reference, signatureType));
				}
			}
			if (this.signatureContent != null) {
				logger.info("[ArangiClientSignatureApplet.coSignReturnStream]::Cofirmando contenido de una firma::" + Arrays.asList(new Object[] { signatureContent, filter, signatureType }));
				if (this.fileToSign != null) {
					logger.info("[ArangiClientSignatureApplet.coSignReturnStream]::Cofirmando documento en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					resultB64 = new SignatureBase64Stream(client.coSign(signatureContent, fileToSign, filter, reference, signatureType));
				} else {
					logger.info("[ArangiClientSignatureApplet.coSignReturnStream]::Cofirmando contenido de documento::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
					resultB64 = new SignatureBase64Stream(client.coSign(signatureContent, contentToSign, filter, reference, signatureType));
				}
			}
			
			//-- Se ha llegado hasta aqu (no se ha definido nada para contrafirmar
			logger.error("[ArangiClientSignatureApplet.coSignReturnStream]::No se ha definido nada que cofirmar");
			throw new NoDocumentToSignException (i18n.getString("error.no.firma.cofirmar"));
		} catch (Throwable e) {
			logger.error ("[ArangiClientSignatureApplet.coSignReturnStream]::ERROR firmando", e);
			this.error = e;
		}
		
	}
	
	/**
	 * Realiza la contrafirma
	 * 
	 * @return Firma en base64
	 */
	public String counterSign () {
		logger.info("[ArangiClientSignatureApplet.counterSign]::Entrada");
		this.error = null;
		
		//-- Pasar las opciones extra
		client.setOpciones(this.options);
				
		try {
			if (this.signatureFile != null) {
				logger.info("[ArangiClientSignatureApplet.counterSign]::Contrafirmando firma en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
				return client.counterSign(signatureFile, filter, signatureType);
			}
			if (this.signatureContent != null) {
				logger.info("[ArangiClientSignatureApplet.counterSign]::Contrafirmando contenido de una firma::" + Arrays.asList(new Object[] { signatureContent, filter, signatureType }));
				return client.counterSign(signatureContent, filter, signatureType);
			}
			
			//-- Se ha llegado hasta aqu (no se ha definido nada para contrafirmar
			logger.error("[ArangiClientSignatureApplet.counterSign]::No se ha definido nada que contrafirmar");
			throw new NoDocumentToSignException (i18n.getString("error.no.firma.contrafirmar"));
		} catch (Throwable e) {
			logger.error ("[ArangiClientSignatureApplet.counterSign]::ERROR firmando", e);
			this.error = e;
			return null;
		}
		
	}

	/**
	 * Realiza la contrafirma. Luego hay que llamar al mtodo getResultPiece para ir obteniendo la firma
	 */
	public void counterSignReturnStream () {
		logger.info("[ArangiClientSignatureApplet.counterSignReturnStream]::Entrada");
		this.error = null;
		
		//-- Pasar las opciones extra
		client.setOpciones(this.options);
				
		try {
			if (this.signatureFile != null) {
				logger.info("[ArangiClientSignatureApplet.counterSignReturnStream]::Contrafirmando firma en fichero::" + Arrays.asList(new Object[] { signatureFile, filter, signatureType }));
				resultB64 = new SignatureBase64Stream(client.counterSign(signatureFile, filter, signatureType));
			}
			if (this.signatureContent != null) {
				logger.info("[ArangiClientSignatureApplet.counterSignReturnStream]::Contrafirmando contenido de una firma::" + Arrays.asList(new Object[] { signatureContent, filter, signatureType }));
				resultB64 = new SignatureBase64Stream(client.counterSign(signatureContent, filter, signatureType));
			}
			
			//-- Se ha llegado hasta aqu (no se ha definido nada para contrafirmar
			logger.error("[ArangiClientSignatureApplet.counterSignReturnStream]::No se ha definido nada que contrafirmar");
			throw new NoDocumentToSignException (i18n.getString("error.no.firma.contrafirmar"));
		} catch (Throwable e) {
			logger.error ("[ArangiClientSignatureApplet.counterSignReturnStream]::ERROR firmando", e);
			this.error = e;
		}
		
	}
	
	/**
	 * Obtiene un trozo del resultado de la firma obtenida con alguno de los mtodos de 
	 * firma. Si devuelve una cadena vaca es que ya ha llegado al final de la firma obtenida.
	 * 
	 * @return Trozo de la firma obtenida
	 */
	public String getResultPiece () {
		logger.info("[ArangiClientSignatureApplet.getResultPiece]::Entrada");
		
		return resultB64.getPiece();
	}

	/**
	 * Muestra el dialogo de ficheros para que el usuario elija uno. 
	 * 
	 * @return Path al fichero elegido
	 */
	public String getPath () {
		logger.info("[ArangiClientSignatureApplet.getPath]::Entrada");
		this.error = null;
		
		//-- Mostrar dilogo para obtener el path
		JFileChooser jfc = new JFileChooser(currentGetPathDirectory);
		jfc.setDialogTitle(i18n.getString("titulo.elegir.path"));
		int ret = jfc.showOpenDialog(null);
		if (ret == JFileChooser.APPROVE_OPTION) {
			String path = jfc.getSelectedFile().getAbsolutePath();
			currentGetPathDirectory = new File (path).getParentFile();
			path = path.replace('\\', '/');
			logger.debug("[ArangiClientSignatureApplet.getPath]::Se devuelve: " + path);
			return path;
		} 
		
		//-- No se devuelve ningn path
		logger.debug("[ArangiClientSignatureApplet.getPath]::No se devuelve ningn path");
		return null;
	}
	

	//-- Mtodos privados
	
	/*
	 * Centra los mensajes en la pantalla
	 */
	private void centerComponent(Component component) {
		Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
		int ancho = component.getSize().width;
		int alto = component.getSize().height;
		int x = (dim.width - ancho) / 2;
		int y = (dim.height - alto) / 2;
		
		// Mueve la ventana a la nueva posicin
		component.setLocation(x, y);
	}

	/**
	 * Obtiene la pila completa de la excepcin
	 */
	private static StringBuffer getFullStackTrace (Throwable exception){ 
		StringBuffer traza = new StringBuffer();
		traza.append("\n\t" + exception.getClass().getName() + ": " + exception.getLocalizedMessage() + "\n");
		StackTraceElement[] st = exception.getStackTrace();
		for (int i = 0; i < st.length; i++) {
			traza.append('\t' + ((StackTraceElement) st[i]).toString() + '\n');
		}
		
		//-- Y las trazas de todas las causas que haya
		while ((exception = (Exception)exception.getCause()) != null){
			traza.append("Caused by: " + exception + ": " + exception.getLocalizedMessage() + "\n");
			st = exception.getStackTrace();
			for (int i = 0; i < st.length; i++) {
				traza.append('\t' + ((StackTraceElement) st[i]).toString() + '\n');
			}
		}
		return traza;
	}
	


}
