package es.accv.arangi.client;

import java.io.File;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Locale;

import org.apache.log4j.Logger;

import es.accv.arangi.base.exception.signature.SignatureNotFoundException;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.client.util.I18n;

/**
 * Applet de validacin del cliente de Arang
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class ArangiClientValidationApplet extends ArangiClientApplet {

	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 1L;
	
	/*
	 * Logger de la clase
	 */
	static Logger logger = Logger.getLogger(ArangiClientValidationApplet.class);
	
	/*
	 * Cliente de Arang
	 */
	private ArangiClient client;
	
	/*
	 * Tipo de firma
	 */
	private String signatureType;
	
	/*
	 * Fichero con la firma a validar
	 */
	private File signatureFile;
	
	/*
	 * Contenido de la firma a validar
	 */
	private byte[] signatureContent;
	
	/*
	 * Fichero que se firm
	 */
	private File detachedFile;
	
	/*
	 * Contenido del fichero que se firm
	 */
	private byte[] detachedContent;
	
	/**
	 * Constructor: carga valores por defecto
	 */
	public ArangiClientValidationApplet () {
		// La firma por defecto es XAdES-BES attached
		this.signatureType = ValidationConstants.SIGNATURE_XADES;
	}
	
	@Override
	public void init() {
		super.init(); 
		
		//-- Obtener los parmetros de pas y lengua
		String language = getParameter("language");
		String country = getParameter("country");
		
		//-- Si no existen, por defecto espaol
		if (language == null) {
			language = "es";
		}
		if (country == null) {
			country = "ES";
		}
		
		//-- Cargar el cliente de Arang y la internacionalizacin
		client = new ArangiClient(language, country);		
		this.i18n = new I18n (new Locale(language, country)); 

	}
	
	//-- Mtodos pblicos
	
	/**
	 * Carga el tipo de firma a utilizar
	 */
	public void setSignatureType (String signatureType) {
		this.signatureType = signatureType;
	}

	/**
	 * Carga el fichero con la firma
	 * 
	 * @param filePath Path al fichero con la firma
	 */
	public void setSignatureFile (String filePath) {
		this.signatureFile = new File (filePath);
		this.signatureContent = null;
	}
	
	/**
	 * Carga el contenido de la firma
	 * 
	 * @param content Contenido de la firma
	 */
	public void setSignatureContent (String content) {
		this.signatureContent = content.getBytes();
		this.signatureFile = null;
	}
	
	/**
	 * Carga el contenido de la firma (el contenido se pasa en base64)
	 * 
	 * @param content Contenido de la firma (se pasa en base64)
	 */
	public void setSignatureContentBase64 (String contentB64) {
		this.signatureContent = Util.decodeBase64(contentB64);
		this.signatureFile = null;
	}
	
	/**
	 * Carga el fichero que se firm
	 * 
	 * @param detachedFilePath Path al fichero que se firm
	 */
	public void setDetachedFile(String detachedFilePath) {
		this.detachedFile = new File (detachedFilePath);
		this.detachedContent = null;
	}

	/**
	 * Carga el contenido del fichero que se firm
	 * 
	 * @param detachedContent Contenido del fichero que se firm
	 */
	public void setDetachedContent(String detachedContent) {
		this.detachedFile = null;
		this.detachedContent = detachedContent.getBytes();
	}

	/**
	 * Carga el contenido del fichero que se firm (el contenido se pasa en base64)
	 * 
	 * @param detachedContent Contenido del fichero que se firm (se pasa en base64)
	 */
	public void setDetachedContentBase64(String detachedContentB64) {
		this.detachedFile = null;
		this.detachedContent = Util.decodeBase64(detachedContentB64);
	}

	/**
	 * Obtiene el texto del error producido
	 * 
	 * @return texto del error
	 */
	public String getError() {
		if (this.error == null) {
			logger.debug("[ArangiClientValidationApplet.getError]::El error es nulo");
			return null;
		}
		
		logger.debug("[ArangiClientValidationApplet.getError]::El error no es nulo", this.error);
		StringWriter writer = new StringWriter();
		PrintWriter pw = new PrintWriter(writer);
		this.error.printStackTrace(pw);
		
		return writer.toString();
	}
	
	/**
	 * Realiza la validacin de la firma
	 * 
	 * @return 
	 */
	public String validateSignature () {
		logger.info("[ArangiClientApplet.validateSignature]::Entrada");
		this.error = null;
		
		try {
			if (this.signatureFile != null) {
				logger.info("[ArangiClientApplet.validateSignature]::Validando una firma en fichero::" + Arrays.asList(new Object[] { signatureFile, signatureType }));
				if (this.detachedFile == null && this.detachedContent == null) {
					return client.getSignatureValidationResultXML(client.validateSignature(signatureFile, signatureType));
				} else {
					if (this.detachedFile != null) {
						return client.getSignatureValidationResultXML (client.validateSignature(signatureFile, detachedFile, signatureType));
					} else {
						return client.getSignatureValidationResultXML (client.validateSignature(signatureFile, detachedContent, signatureType));
					}
				}
			}
			if (this.signatureContent != null) {
				logger.info("[ArangiClientApplet.validateSignature]::Validando el contenido de una firma::" + Arrays.asList(new Object[] { signatureContent, signatureType }));
				if (this.detachedFile == null && this.detachedContent == null) {
					return client.getSignatureValidationResultXML (client.validateSignature(signatureContent, signatureType));
				} else {
					if (this.detachedFile != null) {
						return client.getSignatureValidationResultXML (client.validateSignature(signatureContent, detachedFile, signatureType));
					} else {
						return client.getSignatureValidationResultXML (client.validateSignature(signatureContent, detachedContent, signatureType));
					}
				}
			}
			
			//-- Se ha llegado hasta aqu (no se ha definido nada para firmar
			logger.error("[ArangiClientApplet.validateSignature]::No se ha definido la firma a validar");
			throw new SignatureNotFoundException (i18n.getString("error.no.firma.validar"));
			
		} catch (Throwable e) {
			logger.error ("[ArangiClientApplet.validateSignature]::ERROR validando", e);
			this.error = e;
			return null;
		}
		
	}

	//-- Mtodos privados
	
}
