/**
 * 
 */
package es.accv.arangi.client;

import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

import es.accv.arangi.base.certificate.Certificate;
import es.accv.arangi.base.device.AppleSafariManager;
import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.device.MozillaKeyStoreManager;
import es.accv.arangi.base.device.model.Pkcs11Device;
import es.accv.arangi.base.exception.certificate.NormalizeCertificateException;
import es.accv.arangi.base.exception.device.AppleSafariKeyStoreException;
import es.accv.arangi.base.exception.device.AppleSafariNotFoundException;
import es.accv.arangi.base.exception.device.IAIKDLLNotFoundException;
import es.accv.arangi.base.exception.device.LoadingObjectException;
import es.accv.arangi.base.exception.device.MozillaKeyStoreException;
import es.accv.arangi.base.exception.device.MozillaNotFoundException;
import es.accv.arangi.base.exception.device.OpeningDeviceException;
import es.accv.arangi.base.exception.device.SearchingException;
import es.accv.arangi.base.exception.device.WindowsNotFoundException;
import es.accv.arangi.client.exception.NoCertificateToSignException;
import es.accv.arangi.client.exception.SignatureCancelledByUserException;
import es.accv.arangi.client.filter.ArangiClientFilter;
import es.accv.arangi.client.util.DeviceAndAlias;
import es.accv.arangi.client.util.I18n;
import es.accv.arangi.client.util.PinAndPkcs11Device;
import es.accv.arangi.device.PersonalWindowsKeyStoreManager;
import es.accv.arangi.device.Pkcs11Manager;

/**
 * Clase empleada para obtener un dispositivo y un alias de los presentes en
 * el equipo: en los navegadores o las tarjetas.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class CertificateSelector {

	/*
	 * Constantes para nombres de dispositivos
	 */
	public static final String MOZILLA		= "mozilla";
	public static final String IEXPLORER	= "IExplorer";
	public static final String APPLE		= "Apple";
	public static final String PKCS11		= "PKCS#11";	
	
	/*
	 * Logger de la clase
	 */
	static Logger logger = Logger.getLogger(CertificateSelector.class);
	
	/*
	 * i18n
	 */
	I18n i18n;
	
	/*
	 * Filtro
	 */
	ArangiClientFilter filter;
	
	/*
	 * Flag que indica qu hacer si slo hay un certificado en la seleccin
	 */
	private boolean mostrarSeleccionSiSoloUno	= true;
	
	//-- Constructor
	
	public CertificateSelector (I18n i18n, ArangiClientFilter filter, boolean mostrarSeleccionSiSoloUno) {
		this.i18n = i18n;
		if (filter != null) {
			this.filter = filter;
		} else {
			this.filter = new ArangiClientFilter(); // slo filtra por uso de clave de firmado
		}
		this.mostrarSeleccionSiSoloUno = mostrarSeleccionSiSoloUno;
	}
	
	//-- Mtodos pblicos

	/**
	 * Obtiene un dispositivo y un alias perteneciente a l preguntando al
	 * usuario tras mostrarle todos los certificados que tiene en el ordenador:
	 * navegadores y PKCS#11.
	 * 
	 * @param llamador Componente Swing que llama al mtodo. Se utiliza para poder
	 * abrir dilogos.
	 * @return Dispositivo y alias perteneciente a l
	 * @throws SignatureCancelledByUserException El usuario no desea elegir un certificado
	 * 	de la lista
	 * @throws NoCertificateToSignException Tras aplicar los filtros no hay ningn certificado 
	 *  para firmar
	 */
	public DeviceAndAlias getDeviceAndAlias () throws SignatureCancelledByUserException, NoCertificateToSignException {
		logger.info("[CertificateSelector.getDeviceAndAlias]::Entrada");
		
		HashMap<String, List<DeviceAndAlias>> hmCertificates = new HashMap<String, List<DeviceAndAlias>>();
		
		//-- Obtener la lista de certificados de Firefox
		if (filter.isMozillaKeystore()) {
			logger.debug("[CertificateSelector.getDeviceAndAlias]::Obteniendo certificados de Mozilla");
			MozillaKeyStoreManager mozillaManager = null;
			try {
				//-- Si estamos en Windows slo nos interesan los certificados de Mozilla, pero no los
				//-- de los almacenes PKCS#11 asociados, ya que estos certificados los obtendremos del
				//-- almacn de Windows.
				if (System.getProperty("os.name").contains("indows")) {
					mozillaManager = new MozillaKeyStoreManager(i18n.getString("texto.pin.mozilla"), i18n.getString("titulo.pin.mozilla"), true);
				} else {
					mozillaManager = new MozillaKeyStoreManager(i18n.getString("texto.pin.mozilla"), i18n.getString("titulo.pin.mozilla"), false);
				}
			} catch (MozillaNotFoundException e) {
				logger.info("[CertificateSelector.getDeviceAndAlias]::Mozilla no est instalado en el equipo");
			} catch (MozillaKeyStoreException e) {
				logger.error("[CertificateSelector.getDeviceAndAlias]::Error abriendo el almacn de claves de Mozilla", e);
			}
			List<DeviceAndAlias> lMozillaCertificates = new ArrayList<DeviceAndAlias>();
			if (mozillaManager != null) {
				lMozillaCertificates = getListaCertificados (mozillaManager, MOZILLA, null, this.filter);
			}
			hmCertificates.put(MOZILLA, lMozillaCertificates);
		}
		
		//-- Obtener la lista de certificados de IExplorer
		if (filter.isiExplorerKeystore()) {
			logger.debug("[CertificateSelector.getDeviceAndAlias]::Obteniendo certificados de IExplorer");
			PersonalWindowsKeyStoreManager iExplorerManager = null;
			try {
				iExplorerManager = new PersonalWindowsKeyStoreManager();
			} catch (OpeningDeviceException e) {
				logger.error("[CertificateSelector.getDeviceAndAlias]::Error abriendo el almacn de claves de IExplorer", e);
			} catch (WindowsNotFoundException e) {
				logger.info("[CertificateSelector.getDeviceAndAlias]::La ejecucin no se realiza en Windows");
			}
			List<DeviceAndAlias> lIExplorerCertificates = new ArrayList<DeviceAndAlias>();
			if (iExplorerManager != null) {
				lIExplorerCertificates = getListaCertificados (iExplorerManager, IEXPLORER, null, this.filter);
			}
			hmCertificates.put(IEXPLORER, lIExplorerCertificates);
		}
		
		//-- Obtener los PKCS#11 conectados
		if (filter.isPkcs11Keystore()) {
			List<Pkcs11Device> connectedPkcs11 = null;
			try {
				connectedPkcs11 = Pkcs11Manager.getConnectedDevices();
			} catch (IAIKDLLNotFoundException e) {
				logger.error("[CertificateSelector.getDeviceAndAlias]::No se encuentra la DLL de IAIK", e);
			}
			
			if (connectedPkcs11 != null) {
				HashMap<String, PinAndPkcs11Device> hmPkcs11Data = new HashMap<String, PinAndPkcs11Device>();
				for (Iterator<Pkcs11Device> iterator = connectedPkcs11.iterator(); iterator.hasNext();) {
					Pkcs11Device pkcs11Device = iterator.next();
					logger.debug("[CertificateSelector.getDeviceAndAlias]::Dispositivo PKCS#11 conectado a slot " + pkcs11Device.getToken().getSlot());
					logger.debug("[CertificateSelector.getDeviceAndAlias]::Dispositivo PKCS#11 fabricante" + pkcs11Device.getManufacturerId());
					logger.debug("[CertificateSelector.getDeviceAndAlias]::Dispositivo PKCS#11 modelo" + pkcs11Device.getModel());
					
					
					hmCertificates.put(PKCS11 + "_" + pkcs11Device.getToken().getSlot().getSlotID(), getListaCertificados(pkcs11Device, this.filter));
				}
			}
		}
		
		//-- Ya tenemos los certificados, ahora el usuario debe elegir uno. 
		//-- Primero hacer la lista
		String[] arrayEtiquetasCertificados = getListaEtiquetasCertificados (hmCertificates, filter);
		
		//-- Si el flag indica que si hay slo un certificado no se debe mostrar la seleccin ->
		//-- obtener el nico elemento y devolverlo
		if (arrayEtiquetasCertificados.length == 1 && !mostrarSeleccionSiSoloUno) {
			logger.debug("[CertificateSelector.getDeviceAndAlias]::Como slo hay uno no se muestra la seleccin");
			return getDeviceAndAlias (hmCertificates, arrayEtiquetasCertificados[0]);
		}
		
		//-- Si tras los filtros no hay nada que elegir mostrar un mensaje y lanzar excepcin
		if (arrayEtiquetasCertificados.length == 0) {
			logger.debug("[CertificateSelector.getDeviceAndAlias]::No hay ningn certificado para firmar que cumpla con el filtro establecido");
			JOptionPane.showMessageDialog(null, i18n.getString("error.no.certificado.firma"), 
					i18n.getString("titulo.no.certificado.firma"), JOptionPane.ERROR_MESSAGE);
			throw new NoCertificateToSignException(i18n.getString("error.no.certificado.firma"));	
		}
		
		//-- Mostrar y que elija
		String etiquetaCertificadoElegido;
		while (true) {
			etiquetaCertificadoElegido = (String)JOptionPane.showInputDialog(null, i18n.getString("texto.seleccione.lista.firmar"),
					i18n.getString("titulo.elegir.certificado.firmar"), JOptionPane.PLAIN_MESSAGE, null, arrayEtiquetasCertificados, null);
			logger.debug("[CertificateSelector.getDeviceAndAlias]::El certificado elegido ha sido: " + etiquetaCertificadoElegido);
			
			//-- Si no se ha elegido nada preguntar
			if (etiquetaCertificadoElegido == null) {
				if (JOptionPane.showConfirmDialog(null, i18n.getString("pregunta.no.elegido.certificado"), 
						i18n.getString("pregunta.no.elegido.certificado"), JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {

					logger.debug("[CertificateSelector.getDeviceAndAlias]::El usuario no desea elegir un certificado y continuar con la firma");
					throw new SignatureCancelledByUserException(i18n.getString("error.usuario.cancela.firma"));	
				}
			} else {
				//-- El usuario ha elegido un certificado de la lista
				break;
			}
		}
		
		//-- Buscar el index en la lista del certificado elegido
		return getDeviceAndAlias (hmCertificates, etiquetaCertificadoElegido);
	}
	
	//-- Mtodos privados
	
	/*
	 * Obtiene el deviceandalias seleccionado. Para ello hay que recorrer el hmCertificates
	 * exactamente igual que cuando se cre la lista de etiquetas de certificados.
	 */
	private DeviceAndAlias getDeviceAndAlias( HashMap<String, List<DeviceAndAlias>> hmCertificates, String etiquetaCertificadoElegido) {
		logger.debug("[CertificateSelector.getDeviceAndAlias]::Entrada::" + etiquetaCertificadoElegido);
		
		String tipoKeystore = etiquetaCertificadoElegido.substring(etiquetaCertificadoElegido.indexOf("{") + 1, etiquetaCertificadoElegido.indexOf("}"));
		List<DeviceAndAlias> lCertificates = hmCertificates.get(tipoKeystore);
		
		String commonName = etiquetaCertificadoElegido.substring(etiquetaCertificadoElegido.indexOf(":") + 2, etiquetaCertificadoElegido.indexOf("[") - 1);
		String date = etiquetaCertificadoElegido.substring(etiquetaCertificadoElegido.indexOf("[") + 1, etiquetaCertificadoElegido.indexOf("]"));
		int posicionDeBusqueda = Integer.parseInt (etiquetaCertificadoElegido.substring(etiquetaCertificadoElegido.lastIndexOf("[") + 1, 
				etiquetaCertificadoElegido.lastIndexOf("]")));
		
		int posicion = 1;
		for (Iterator iterator = lCertificates.iterator(); iterator.hasNext();) {
			DeviceAndAlias deviceAndAlias = (DeviceAndAlias) iterator.next();
			if (deviceAndAlias.getCertificate().getCommonName().equals(commonName) &&
					ArangiClient.DATE_FORMAT.format (deviceAndAlias.getCertificate().getValidityPeriodEnd()).equals(date)) {
				if (posicion == posicionDeBusqueda) {
					logger.debug("[CertificateSelector.getDeviceAndAlias]::Elemento encontrado::Keystore: " + deviceAndAlias.getKeystoreType());
					logger.debug("[CertificateSelector.getDeviceAndAlias]::Elemento encontrado::Alias: " + deviceAndAlias.getAlias());
					logger.debug("[CertificateSelector.getDeviceAndAlias]::Elemento encontrado::Certificado: " + deviceAndAlias.getCertificate().getCommonName());
					return deviceAndAlias;
				} else {
					posicion++;
				}
			}
		}
		
		logger.debug("[CertificateSelector.getDeviceAndAlias]::No se ha encontrado el elemento");
		return null;
	}

	/*
	 * Obtiene una lista de certificados de un dispositivo
	 */
	private static List<DeviceAndAlias> getListaCertificados (DeviceManager deviceManager, String keystoreType, String pin, ArangiClientFilter filter) {
		List<DeviceAndAlias> lCertificates = new ArrayList<DeviceAndAlias>();
		String[] alias = null;
		try {
			alias = deviceManager.getAliasNamesList();
		} catch (SearchingException e) {
			logger.error("[CertificateSelector.getListaCertificados]::No se puede consulta la lista de " +
					"alias del dispositivo", e);
		}
		if (alias != null) {
			for (int i = 0; i < alias.length; i++) {
				try {
					logger.debug("[CertificateSelector.getListaCertificados]::Comprobar si el alias '" + alias[i] + "' tiene clave privada");
					try {
						if (deviceManager.getPrivateKey(alias[i]) == null) {
							logger.debug("[CertificateSelector.getListaCertificados]::El alias '" + alias[i] + "' no tiene clave privada");
							continue;
						}
					} catch (LoadingObjectException e) {
						logger.debug("[CertificateSelector.getListaCertificados]::No se puede comprobar si el alias '" + alias[i] + "' tiene clave privada");
						continue;
					}
					Certificate certificate = new Certificate (deviceManager.getCertificate(alias[i]));
					if (filter.matchFilter(certificate)) {
						lCertificates.add(new DeviceAndAlias (keystoreType, alias[i], null, certificate, pin));
					}
				} catch (NormalizeCertificateException e) {
					logger.error("[CertificateSelector.getListaCertificados]::No se puede normalizar el certificado que se " +
							"encuentra en el alias '" + alias[i] + "'", e);
				} catch (SearchingException e) {
					logger.error("[CertificateSelector.getListaCertificados]::No se puede obtener el certificado que se  " +
							"encuentra en el alias '" + alias[i] + "'", e);
				}
			}
		}
		
		return lCertificates;
	}
	
	/*
	 * Obtiene una lista ordenada de etiquetas de certificados siguiendo el siguiente orden:
	 * primero Mozilla, luego IExplorer y al final los PKCS11 en orden de slot.
	 */
	private String[] getListaEtiquetasCertificados (HashMap<String, List<DeviceAndAlias>> hmCertificates, ArangiClientFilter filter) {
		
		ArrayList<String> lEtiquetasCertificados = new ArrayList<String>();
		HashSet<String> setSKIs = new HashSet<String>();
		HashMap<String,Integer> hmNumeraciones = new HashMap<String, Integer>();

		if (filter.isAppleKeystore()) {
			logger.debug("[CertificateSelector.getListaEtiquetasCertificados]::Etiquetas para Apple Safari");
			if (hmCertificates.containsKey(CertificateSelector.APPLE)) {
				List<DeviceAndAlias> lCertificates = hmCertificates.get(CertificateSelector.APPLE);
				for (Iterator<DeviceAndAlias> iterator = lCertificates.iterator(); iterator.hasNext();) {
					Certificate certificate = iterator.next().getCertificate();
					if (!setSKIs.contains(certificate.getSubjectKeyIdentifier())) {
						String etiqueta = "{" + CertificateSelector.APPLE + "}: " + certificate.getCommonName() + 
							" [" + ArangiClient.DATE_FORMAT.format(certificate.getValidityPeriodEnd()) + "]";
						int posicion = 1;
						if (hmNumeraciones.containsKey(etiqueta)) {
							posicion = hmNumeraciones.get(etiqueta);
						}
						lEtiquetasCertificados.add(etiqueta + " [" + posicion + "]");
						hmNumeraciones.put(etiqueta, posicion + 1);
					} else {
						logger.info("[CertificateSelector.getListaEtiquetasCertificados]::No se ha aadido el certificado porque ya existe en otro keystore: " + certificate.getCommonName());
					}
					setSKIs.add(certificate.getSubjectKeyIdentifier());
				}
			}
		}
		
		if (filter.isMozillaKeystore()) {
			logger.debug("[CertificateSelector.getListaEtiquetasCertificados]::Etiquetas para Mozilla");
			if (hmCertificates.containsKey(CertificateSelector.MOZILLA)) {
				List<DeviceAndAlias> lCertificates = hmCertificates.get(CertificateSelector.MOZILLA);
				for (Iterator<DeviceAndAlias> iterator = lCertificates.iterator(); iterator.hasNext();) {
					Certificate certificate = iterator.next().getCertificate();
					if (!setSKIs.contains(certificate.getSubjectKeyIdentifier())) {
						String etiqueta = "{" + CertificateSelector.MOZILLA + "}: " + certificate.getCommonName() + 
							" [" + ArangiClient.DATE_FORMAT.format(certificate.getValidityPeriodEnd()) + "]";
						int posicion = 1;
						if (hmNumeraciones.containsKey(etiqueta)) {
							posicion = hmNumeraciones.get(etiqueta);
						}
						lEtiquetasCertificados.add(etiqueta + " [" + posicion + "]");
						hmNumeraciones.put(etiqueta, posicion + 1);
					} else {
						logger.info("[CertificateSelector.getListaEtiquetasCertificados]::No se ha aadido el certificado porque ya existe en otro keystore: " + certificate.getCommonName());
					}
					setSKIs.add(certificate.getSubjectKeyIdentifier());
				}
			}
		}
		
		if (filter.isiExplorerKeystore()) {
			logger.debug("[CertificateSelector.getListaEtiquetasCertificados]::Etiquetas para IExplorer");
			if (hmCertificates.containsKey(CertificateSelector.IEXPLORER)) {
				List<DeviceAndAlias> lCertificates = hmCertificates.get(CertificateSelector.IEXPLORER);
				for (Iterator<DeviceAndAlias> iterator = lCertificates.iterator(); iterator.hasNext();) {
					Certificate certificate = iterator.next().getCertificate();
					if (!setSKIs.contains(certificate.getSubjectKeyIdentifier())) {
						String etiqueta = "{" + CertificateSelector.IEXPLORER + "}: " + certificate.getCommonName() + 
							" [" + ArangiClient.DATE_FORMAT.format(certificate.getValidityPeriodEnd()) + "]";
						int posicion = 1;
						if (hmNumeraciones.containsKey(etiqueta)) {
							posicion = hmNumeraciones.get(etiqueta);
						}
						lEtiquetasCertificados.add(etiqueta + " [" + posicion + "]");
						hmNumeraciones.put(etiqueta, posicion + 1);
					} else {
						logger.info("[CertificateSelector.getListaEtiquetasCertificados]::No se ha aadido el certificado porque ya existe en otro keystore: " + certificate.getCommonName());
					}
					setSKIs.add(certificate.getSubjectKeyIdentifier());
				}
			}
		}
		
		if (filter.isPkcs11Keystore()) {
			//-- Obtener los nombres de los PKCS#11 y ordenarlos alfabticamente
			logger.debug("[CertificateSelector.getListaEtiquetasCertificados]::Etiquetas para PKCS#11");
			ArrayList<String> pkcs11KeystoreNames = new ArrayList<String>();
			for (Iterator<String> iterator = hmCertificates.keySet().iterator(); iterator.hasNext();) {
				String keystoreName = iterator.next();
				if (keystoreName.startsWith(CertificateSelector.PKCS11)) {
					pkcs11KeystoreNames.add(keystoreName);
				}
			}
			Collections.sort(pkcs11KeystoreNames);
			
			for (Iterator<String> iteratorkeystores = pkcs11KeystoreNames.iterator(); iteratorkeystores.hasNext();) {
				String keystoreName = iteratorkeystores.next();
				List<DeviceAndAlias> lCertificates = hmCertificates.get(keystoreName);
				for (Iterator<DeviceAndAlias> iterator = lCertificates.iterator(); iterator.hasNext();) {
					Certificate certificate = iterator.next().getCertificate();
					if (!setSKIs.contains(certificate.getSubjectKeyIdentifier())) {
						String etiqueta = "{" + keystoreName + "}: " + certificate.getCommonName() + 
							" [" + ArangiClient.DATE_FORMAT.format(certificate.getValidityPeriodEnd()) + "]";
						int posicion = 1;
						if (hmNumeraciones.containsKey(etiqueta)) {
							posicion = hmNumeraciones.get(etiqueta);
						}
						lEtiquetasCertificados.add(etiqueta + " [" + posicion + "]");
						hmNumeraciones.put(etiqueta, posicion + 1);
					} else {
						logger.info("[CertificateSelector.getListaEtiquetasCertificados]::No se ha aadido el certificado porque ya existe en otro keystore: " + certificate.getCommonName());
					}
					setSKIs.add(certificate.getSubjectKeyIdentifier());
				}
			}
		}
		
		return lEtiquetasCertificados.toArray(new String[0]);
	}
	
	/*
	 * Obtiene los certificados de un keystore pkcs#11
	 */
	private static List<DeviceAndAlias> getListaCertificados (Pkcs11Device pkcs11Device, ArangiClientFilter filter) {
		logger.debug ("[Util.getArangiCertificates]::Entrada");
		
		X509Certificate[] arrayX509Certificates = pkcs11Device.getCertificates();
		String[] arrayAlias = pkcs11Device.getAliases();
		
		if (arrayX509Certificates == null) {
			arrayX509Certificates = new X509Certificate [0];
		}
		
		List<DeviceAndAlias> lCertificates = new ArrayList<DeviceAndAlias>();
		for (int i = 0; i < arrayX509Certificates.length; i++) {
			try {
				Certificate certificate = new Certificate (arrayX509Certificates[i]);
				if (filter.matchFilter(certificate)) {
					lCertificates.add(new DeviceAndAlias (PKCS11, arrayAlias[i], pkcs11Device, certificate, null));
				}
			} catch (NormalizeCertificateException e) {
				logger.error("[CertificateSelector.toArangiCertificates]::No se puede normalizar el certificado" + arrayX509Certificates[i], e);
			}
		}
		
		return lCertificates;
	}

	

}
