package es.accv.arangi.client.filter;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;

import es.accv.arangi.base.certificate.Certificate;

/**
 * Filtro para seleccionar certificados. Se puede filtrar por
 * varios campos del certificado e incluso elegir de qu 
 * keystores seleccionar.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class ArangiClientFilter {

	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(ArangiClientFilter.class);
	
	/*
	 * Filtro de keystore Firefox
	 */
	boolean mozillaKeystore = true;
	
	/*
	 * Filtro de keystore IExplorer
	 */
	boolean iExplorerKeystore = true;
	
	/*
	 * Filtro de keystore Apple Safari
	 */
	boolean appleKeystore = true;
	
	/*
	 * Filtro de keystore Pkcs11
	 */
	boolean pkcs11Keystore = true;
	
	/*
	 * Filtro por subject DN
	 */
	List<String> subjectDNFilter = new ArrayList<String>();
	
	/*
	 * Filtro por issuer DN
	 */
	List<String> issuerDNFilter = new ArrayList<String>();
	
	/*
	 * Filtro por serial number
	 */
	List<String> serialNumberFilter = new ArrayList<String>();
	
	/*
	 * Filtro todos los usos de clave. Si no slo se filtran los de firma.
	 */
	boolean allKeyUsages;
	
	//-- Mtodos pblicos
	
	public boolean matchFilter (Certificate certificate) {
		
		if (certificate == null) {
			return false;
		}
		
		logger.info("[ArangiClientFilter.filter]::matchFilter::" + certificate.getCommonName());
		
		//-- Aplicar el filtro de los usos de firma
		if (!allKeyUsages) {
			logger.debug("[ArangiClientFilter.filter]::Filtrando slo los certificados de firma");
			if (!certificate.isKeyUsageDigitalSignature() && !certificate.isKeyUsageNonRepudiation()) {
				logger.debug("[ArangiClientFilter.filter]::No es un certificado de firma");
				return false;
			}
		}
		
		//-- Aplicar filtro del subject
		logger.debug("[ArangiClientFilter.filter]::Filtrando por Subject DN: " + subjectDNFilter);
		if (!filterText(subjectDNFilter, certificate.getSubjectDN())) {
			return false;
		}
		
		//-- Aplicar filtro del issuer
		logger.debug("[ArangiClientFilter.filter]::Filtrando por Issuer DN: " + issuerDNFilter);
		if (!filterText(issuerDNFilter, certificate.getIssuerDN())) {
			return false;
		}

		//-- Aplicar filtro del nmero de serie
		logger.debug("[ArangiClientFilter.filter]::Filtrando por Nmero de Serie: " + serialNumberFilter);
		if (!filterText(serialNumberFilter, certificate.getSerialNumber())) {
			return false;
		}
		
		//-- Ha pasado todos los filtros
		logger.debug("[ArangiClientFilter.filter]::El certificado '" + certificate.getCommonName() + "' ha pasado todos los filtros");
		return true;		
	}
	
	//-- Mtodos privados
	
	/*
	 * Filtra un texto en base a una serie de filtros
	 */
	private boolean filterText (List<String> filtro, String textoAFiltrar) {
		if (!filtro.isEmpty()) {
			boolean filtroTotalPasado = false;
			for (Iterator<String> iterator = filtro.iterator(); iterator.hasNext();) {
				String rule = iterator.next();
				logger.debug("[ArangiClientFilter.filterText]::Filtrando por: " + rule);
				boolean filtroPasado = true;
				StringTokenizer st = new StringTokenizer(rule, ",");
				while (st.hasMoreTokens()) {
					String token = st.nextToken();
					if (textoAFiltrar.indexOf(token) == -1) {
						filtroPasado = false;
						break;
					}
				}
				if (filtroPasado) {
					//-- Se ha pasado uno de los filtros, as que se da por bueno
					logger.debug("[ArangiClientFilter.filterText]::Hace match con la regla: " + rule);
					filtroTotalPasado = true;
					break;
				} else {
					logger.debug("[ArangiClientFilter.filterText]::No hace match con la regla: " + rule);
				}
			}
			
			//-- Si no se ha pasado ningn filtro se da por malo
			if (!filtroTotalPasado) {
				return false;
			}
			
		}
		return true;

	}
	
	//-- GET / SET

	public void addSubjectDNFilterRule(String rule) {
		this.subjectDNFilter.add(rule);
	}

	public void addIssuerDNFilterRule(String rule) {
		this.issuerDNFilter.add(rule);
	}

	public void addSerialNumberFilterRule(String rule) {
		this.serialNumberFilter.add(rule);
	}

	public void setAllKeyUsages(boolean allKeyUsages) {
		this.allKeyUsages = allKeyUsages;
	}

	public boolean isiExplorerKeystore() {
		return iExplorerKeystore;
	}

	public void setIExplorerKeystore(boolean iExplorerKeystore) {
		this.iExplorerKeystore = iExplorerKeystore;
	}

	public boolean isPkcs11Keystore() {
		return pkcs11Keystore;
	}

	public boolean isAppleKeystore() {
		return appleKeystore;
	}

	public void setAppleKeystore(boolean appleKeystore) {
		this.appleKeystore = appleKeystore;
	}
	
	public void setPkcs11Keystore(boolean pkcs11Keystore) {
		this.pkcs11Keystore = pkcs11Keystore;
	}

	public boolean isMozillaKeystore() {
		return mozillaKeystore;
	}

	public void setMozillaKeystore(boolean mozillaKeystore) {
		this.mozillaKeystore = mozillaKeystore;
	}

	

	
	
	
	
}
