/**
 * 
 */
package es.accv.arangi.client.signature;

import java.io.File;
import java.io.InputStream;
import java.util.Map;

import es.accv.arangi.base.certificate.Certificate;
import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.exception.signature.SignatureNotFoundException;
import es.accv.arangi.base.signature.ISignature;

/**
 * Interfaz que han de definir las clases que construyen los diferentes tipos
 * de firma del cliente de Arang.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public interface ArangiClientSignatureFactory {

	/**
	 * Firma en fichero
	 * 
	 * @param file Fichero a firmar
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws NoDocumentToSignException El fichero a firmar no existe o no se puede leer
	 * @throws SignatureException Error durante el proceso de firma
	 */
	public String sign(File file, DeviceManager manager, String alias, String reference) throws NoDocumentToSignException, SignatureException;

	/**
	 * Firma el contenido pasado como parmetro
	 * 
	 * @param contentToSign
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de firma
	 */
	public String sign(byte[] contentToSign, DeviceManager manager, String alias, String reference) throws SignatureException;

	/**
	 * Realiza una cofirma a partir de la firma y el documento en fichero
	 * 
	 * @param signature Fichero con la firma
	 * @param document Documento que origin la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 * @throws SignatureNotFoundException El fichero de la firma no existe o no puede ser
	 * 	ledo
	 */
	public String coSign(File signature, File document, DeviceManager manager, String alias, String reference) throws SignatureException, SignatureNotFoundException;

	/**
	 * Realiza una cofirma a partir de los contenidos de la firma y el documento
	 * 
	 * @param signatureContent Contenido de la firma
	 * @param documentContent Contenido del documento que origin la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 */
	public String coSign(byte[] signatureContent, byte[] documentContent, DeviceManager manager, String alias, String reference) throws SignatureException;

	/**
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param signature Firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de contrafirma
	 */
	public String counterSign(byte[] signature, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException;
	
	/**
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param signatureFile Fichero con la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de contrafirma
	 */
	public String counterSign(File signatureFile, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException;
	
	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un fichero.
	 * 
	 * @param signature Fichero con la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(File signature) throws Exception;

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un array de bytes.
	 * 
	 * @param signature Contenido de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(byte[] signature) throws Exception;

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un stream de lectura.
	 * 
	 * @param isSignature Stream al fichero de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(InputStream isSignature) throws Exception;

	/**
	 * Pasa a la factory una serie de opciones que pueden variar
	 * su comportamiento durante la firma.
	 * 
	 * @param options Opciones
	 */
	public void setSignatureOptions (Map<String, Object> options);
	
	/**
	 * Cambia el valor de la TSA a utilizar
	 */
	public void setUrlTSA (String urlTSA);

}
