package es.accv.arangi.client.signature;

import java.io.File;
import java.io.InputStream;
import java.util.Arrays;

import org.apache.log4j.Logger;

import es.accv.arangi.base.algorithm.DigitalSignatureAlgorithm;
import es.accv.arangi.base.certificate.Certificate;
import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.document.ByteArrayDocument;
import es.accv.arangi.base.document.FileDocument;
import es.accv.arangi.base.exception.device.AliasNotFoundException;
import es.accv.arangi.base.exception.device.LoadingObjectException;
import es.accv.arangi.base.exception.device.SearchingException;
import es.accv.arangi.base.exception.document.HashingException;
import es.accv.arangi.base.exception.document.InitDocumentException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.exception.signature.SignatureNotFoundException;
import es.accv.arangi.base.signature.CMSSignature;
import es.accv.arangi.base.signature.ISignature;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.client.util.I18n;

/**
 * Clase que realiza firmas CMS
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class CmsSignatureFactory  extends BaseSignatureFactory {
	
	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(CmsSignatureFactory.class);
	
	/*
	 * i18n
	 */
	I18n i18n;
	
	/**
	 * Constructor de todas las factory de firma
	 * 
	 * @param attached La firma es attached
	 * @param i18n Internacionalizacin
	 */
	public CmsSignatureFactory (Boolean attached, I18n i18n) {
		this.i18n = i18n;
	}

	/* (non-Javadoc)
	 * @see es.accv.arangi.client.signature.ArangiClientSignatureFactory#sign(java.io.File, es.accv.arangi.base.device.DeviceManager, java.lang.String)
	 */
	public String sign(File file, DeviceManager manager, String alias, String reference) throws NoDocumentToSignException, SignatureException {
		logger.info("[CmsSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { file, manager, alias, reference }));
		
		//-- Obtener el documento
		FileDocument document;
		try {
			document = new FileDocument(file);
		} catch (InitDocumentException e) {
			logger.error("[CmsSignatureFactory.sign]::No ha sido posible leer el fichero " + file, e);
			throw new NoDocumentToSignException(i18n.getString("error.no.documento.leer", file), e);
		}
		
		//-- Firmar
		logger.debug("[CmsSignatureFactory.sign]::Firmando...");
		CMSSignature signature;
		try {
			signature = CMSSignature.sign(new DeviceManager[] { manager }, new String[] { alias }, document, new String[] { DigitalSignatureAlgorithm.getDefault() }, false);
		} catch (AliasNotFoundException e) {
			logger.error("[CmsSignatureFactory.sign]::No se encuentra el alias '" + alias + "' en el dispositivo", e);
			throw new SignatureException(i18n.getString("error.no.alias.dispositivo", alias), e);
		} catch (HashingException e) {
			logger.error("[CmsSignatureFactory.sign]::Error obteniendo el hash del documento", e);
			throw new SignatureException(i18n.getString("error.hash.documento"), e);
		} catch (LoadingObjectException e) {
			logger.error("[CmsSignatureFactory.sign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[CmsSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} catch (SearchingException e) {
			logger.error("[CmsSignatureFactory.sign]::No es posible obtener el alias '" + alias + "' en el dispositivo", e);
			throw new SignatureException(i18n.getString("error.obteniendo.alias", alias), e);
		}
		
		logger.debug("[CmsSignatureFactory.sign]::Se ha obtenido la firma");
		return Util.encodeBase64(signature.toByteArray());
	}

	public String sign(byte[] contentToSign, DeviceManager manager, String alias, String reference) throws SignatureException {
		logger.info("[CmsSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { contentToSign, manager, alias, reference }));
		
		//-- Obtener el documento
		ByteArrayDocument document = new ByteArrayDocument(contentToSign);
		
		//-- Firmar
		logger.debug("[CmsSignatureFactory.sign]::Firmando...");
		CMSSignature signature;
		try {
			signature = CMSSignature.sign(new DeviceManager[] { manager }, new String[] { alias }, document, new String[] { DigitalSignatureAlgorithm.getDefault() }, false);
		} catch (AliasNotFoundException e) {
			logger.error("[CmsSignatureFactory.sign]::No se encuentra el alias '" + alias + "' en el dispositivo", e);
			throw new SignatureException(i18n.getString("error.no.alias.dispositivo", alias), e);
		} catch (HashingException e) {
			logger.error("[CmsSignatureFactory.sign]::Error obteniendo el hash del documento", e);
			throw new SignatureException(i18n.getString("error.hash.documento"), e);
		} catch (LoadingObjectException e) {
			logger.error("[CmsSignatureFactory.sign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[CmsSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} catch (SearchingException e) {
			logger.error("[CmsSignatureFactory.sign]::No es posible obtener el alias '" + alias + "' en el dispositivo", e);
			throw new SignatureException(i18n.getString("error.obteniendo.alias", alias), e);
		}
		
		logger.debug("[CmsSignatureFactory.sign]::Se ha obtenido la firma");
		return Util.encodeBase64(signature.toByteArray());
	}

	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una cofirma a partir de la firma y el documento en fichero
	 * 
	 * @param fSignature Fichero con la firma
	 * @param fDocument Documento que origin la firma (puede ser nulo)
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 * @throws SignatureNotFoundException El fichero de la firma no existe o no puede ser
	 * 	ledo
	 */
	public String coSign(File fSignature, File fDocument, DeviceManager manager, String alias, String reference) throws SignatureException, SignatureNotFoundException {
		logger.info("[Pkcs7SignatureFactory.coSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}

	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una cofirma a partir de los contenidos de la firma y el documento
	 * 
	 * @param signatureContent Contenido de la firma
	 * @param documentContent Contenido del documento que origin la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 */
	public String coSign(byte[] signatureContent, byte[] documentContent, DeviceManager manager, String alias, String reference) throws SignatureException {
		logger.info("[Pkcs7SignatureFactory.coSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}

	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param signature Firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return
	 * @throws SignatureException
	 */
	public String counterSign(byte[] signature, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException {
		logger.info("[CmsSignatureFactory.counterSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}
	
	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param signatureFile Fichero con la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return
	 * @throws SignatureException
	 */
	public String counterSign(File signatureFile, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException {
		logger.info("[CmsSignatureFactory.counterSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}
	
	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un fichero.
	 * 
	 * @param signature Fichero con la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(File signature) throws Exception {
		return new CMSSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un array de bytes.
	 * 
	 * @param signature Contenido de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(byte[] signature) throws Exception {
		return new CMSSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un stream de lectura.
	 * 
	 * @param isSignature Stream al fichero de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(InputStream isSignature) throws Exception {
		return new CMSSignature(isSignature);
	}

}
