package es.accv.arangi.client.signature;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Arrays;

import org.apache.log4j.Logger;

import es.accv.arangi.base.certificate.Certificate;
import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.document.ByteArrayDocument;
import es.accv.arangi.base.document.FileDocument;
import es.accv.arangi.base.exception.certificate.CertificateCANotFoundException;
import es.accv.arangi.base.exception.device.AliasNotFoundException;
import es.accv.arangi.base.exception.device.LoadingObjectException;
import es.accv.arangi.base.exception.document.HashingException;
import es.accv.arangi.base.exception.document.InitDocumentException;
import es.accv.arangi.base.exception.signature.InvalidCertificateException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.PDFDocumentException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.exception.signature.SignatureNotFoundException;
import es.accv.arangi.base.signature.ISignature;
import es.accv.arangi.base.signature.PDFSignature;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.client.ArangiClient;
import es.accv.arangi.client.util.I18n;
import es.accv.arangi.timestamp.TimeStamp;

/**
 * Clase que realiza firmas PDF con sello de tiempos
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class PdfSignatureFactory extends BaseSignatureFactory {

	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(PdfSignatureFactory.class);
	
	/*
	 * i18n
	 */
	I18n i18n;
	
	/**
	 * Constructor de todas las factory de firma
	 * 
	 * @param attached La firma es attached
	 * @param i18n Internacionalizacin
	 */
	public PdfSignatureFactory (Boolean attached, I18n i18n) {
		this.i18n = i18n;
	}

	/* (non-Javadoc)
	 * @see es.accv.arangi.client.signature.ArangiClientSignatureFactory#sign(java.io.File, es.accv.arangi.base.device.DeviceManager, java.lang.String)
	 */
	public String sign(File file, DeviceManager manager, String alias, String reference) throws NoDocumentToSignException, SignatureException {
		logger.info("[PdfSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { file, manager, alias, reference }));
		
		//-- Obtener el documento
		FileDocument document;
		try {
			document = new FileDocument(file);
		} catch (InitDocumentException e) {
			logger.error("[PdfSignatureFactory.sign]::No ha sido posible leer el fichero " + file, e);
			throw new NoDocumentToSignException(i18n.getString("error.no.documento.leer", file), e);
		}
		
		//-- Ver si existe la opcin de firmar sin sello
		boolean firmarSinSello = options.get(ArangiClient.OPCION_FIRMAR_SIN_SELLO) != null && options.get(ArangiClient.OPCION_FIRMAR_SIN_SELLO) instanceof Boolean &&
				((Boolean) options.get(ArangiClient.OPCION_FIRMAR_SIN_SELLO)).booleanValue();
		URL urlTimestamp = TimeStamp.getURLACCVTSA();
		if (firmarSinSello) {
			urlTimestamp = null;
		}
		
		//-- firma visible?
		String pathImagen = (String) options.get(ArangiClient.OPCION_IMAGE_PATH);
		byte[] contenidoImagen = null;
		if (pathImagen != null) {
			try {
				contenidoImagen = Util.loadFile(new File (pathImagen));
			} catch (IOException e) {
				logger.debug("[PdfSignatureFactory.sign]::No se puede cargar la imagen en '" + pathImagen + "'");
			}
		}
		Float xInferiorIzquierda = null;
		try {
			xInferiorIzquierda = Float.parseFloat((String) options.get(ArangiClient.OPCION_X_BOTTOM_LEFT));
		} catch (Exception e) {};
		Float yInferiorIzquierda = null;
		try {
			yInferiorIzquierda = Float.parseFloat((String) options.get(ArangiClient.OPCION_Y_BOTTOM_LEFT));
		} catch (Exception e) {};
		Float xSuperiorDerecha = null;
		try {
			xSuperiorDerecha = Float.parseFloat((String) options.get(ArangiClient.OPCION_X_TOP_RIGHT));
		} catch (Exception e) {};
		Float ySuperiorDerecha = null;
		try {
			ySuperiorDerecha = Float.parseFloat((String) options.get(ArangiClient.OPCION_Y_TOP_RIGHT));
		} catch (Exception e) {};
		Integer pagina = null;
		try {
			pagina = Integer.parseInt((String) options.get(ArangiClient.OPCION_PAGE));
		} catch (Exception e) {};
		boolean firmaVisible = false;
		if (xInferiorIzquierda != null && yInferiorIzquierda != null &&
				xSuperiorDerecha != null && ySuperiorDerecha != null && pagina != null) {
			firmaVisible = true;
		}
		
		//-- Firmar
		logger.debug("[PdfSignatureFactory.sign]::Firmando...");
		PDFSignature signature;
		try {
			if (firmaVisible) {
				signature = PDFSignature.sign(new DeviceManager[] { manager }, new String[] { alias }, document, urlTimestamp, null, null, null, reference!=null?reference:i18n.getString("pdf.signature.reason"), 
						true, contenidoImagen, xInferiorIzquierda, yInferiorIzquierda, xSuperiorDerecha, ySuperiorDerecha, pagina);
			} else {
				signature = PDFSignature.sign(new DeviceManager[] { manager }, new String[] { alias }, document, urlTimestamp, null, reference!=null?reference:i18n.getString("pdf.signature.reason"));
			}
		} catch (AliasNotFoundException e) {
			logger.error("[PdfSignatureFactory.sign]::No se encuentra el alias '" + alias + "' en el dispositivo", e);
			throw new SignatureException(i18n.getString("error.no.alias.dispositivo", alias), e);
		} catch (HashingException e) {
			logger.error("[PdfSignatureFactory.sign]::Error obteniendo el hash del documento", e);
			throw new SignatureException(i18n.getString("error.hash.documento"), e);
		} catch (LoadingObjectException e) {
			logger.error("[PdfSignatureFactory.sign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[PdfSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} catch (PDFDocumentException e) {
			logger.error("[PdfSignatureFactory.sign]::El documento no es un PDF o es un PDF mal formado", e);
			throw new SignatureException(i18n.getString("error.pdf.mal.formado"), e);
		} catch (CertificateCANotFoundException e) {
			logger.error("[PdfSignatureFactory.sign]::El certificado no pertenece a una de las autoridades de certificacin de confianza", e);
			throw new SignatureException(i18n.getString("error.no.ca.confianza"), e);
		} catch (InvalidCertificateException e) {
			logger.error("[PdfSignatureFactory.sign]::El certificado con el que se pretende firmar est revocado", e);
			throw new SignatureException(i18n.getString("error.firmando.certificado.revocado"), e);
		} 
		
		logger.debug("[PdfSignatureFactory.sign]::Se ha obtenido la firma");
		try {
			return Util.encodeBase64(signature.toByteArray());
		} catch (IOException e) {
			logger.error("[PdfSignatureFactory.sign]::No se ha podido leer el fichero PDF firmado", e);
			throw new SignatureException(i18n.getString("error.lectura.pdf.firmado"), e);
		}
	}

	public String sign(byte[] contentToSign, DeviceManager manager, String alias, String reference) throws SignatureException {
		logger.info("[PdfSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { contentToSign, manager, alias, reference }));
		
		//-- Obtener el documento
		ByteArrayDocument document = new ByteArrayDocument(contentToSign);
		
		//-- Ver si existe la opcin de firmar sin sello
		boolean firmarSinSello = options.get(ArangiClient.OPCION_FIRMAR_SIN_SELLO) != null && options.get(ArangiClient.OPCION_FIRMAR_SIN_SELLO) instanceof Boolean &&
				((Boolean) options.get(ArangiClient.OPCION_FIRMAR_SIN_SELLO)).booleanValue();
		URL urlTimestamp = TimeStamp.getURLACCVTSA();
		if (firmarSinSello) {
			urlTimestamp = null;
		}
		
		//-- firma visible?
		String pathImagen = (String) options.get(ArangiClient.OPCION_IMAGE_PATH);
		byte[] contenidoImagen = null;
		if (pathImagen != null) {
			try {
				contenidoImagen = Util.loadFile(new File (pathImagen));
			} catch (IOException e) {
				logger.debug("[PdfSignatureFactory.sign]::No se puede cargar la imagen en '" + pathImagen + "'");
			}
		}
		Float xInferiorIzquierda = null;
		try {
			xInferiorIzquierda = Float.parseFloat((String) options.get(ArangiClient.OPCION_X_BOTTOM_LEFT));
		} catch (Exception e) {};
		Float yInferiorIzquierda = null;
		try {
			yInferiorIzquierda = Float.parseFloat((String) options.get(ArangiClient.OPCION_Y_BOTTOM_LEFT));
		} catch (Exception e) {};
		Float xSuperiorDerecha = null;
		try {
			xSuperiorDerecha = Float.parseFloat((String) options.get(ArangiClient.OPCION_X_TOP_RIGHT));
		} catch (Exception e) {};
		Float ySuperiorDerecha = null;
		try {
			ySuperiorDerecha = Float.parseFloat((String) options.get(ArangiClient.OPCION_Y_TOP_RIGHT));
		} catch (Exception e) {};
		Integer pagina = null;
		try {
			pagina = Integer.parseInt((String) options.get(ArangiClient.OPCION_PAGE));
		} catch (Exception e) {};
		boolean firmaVisible = false;
		if (xInferiorIzquierda != null && yInferiorIzquierda != null &&
				xSuperiorDerecha != null && ySuperiorDerecha != null && pagina != null) {
			firmaVisible = true;
		}
		
		//-- Firmar
		logger.debug("[PdfSignatureFactory.sign]::Firmando...");
		PDFSignature signature;
		try {
			if (firmaVisible) {
				signature = PDFSignature.sign(new DeviceManager[] { manager }, new String[] { alias }, document, urlTimestamp, null, null, null, reference!=null?reference:i18n.getString("pdf.signature.reason"), 
						true, contenidoImagen, xInferiorIzquierda, yInferiorIzquierda, xSuperiorDerecha, ySuperiorDerecha, pagina);
			} else {
				signature = PDFSignature.sign(new DeviceManager[] { manager }, new String[] { alias }, document, urlTimestamp, null, reference!=null?reference:i18n.getString("pdf.signature.reason"));
			}
		} catch (AliasNotFoundException e) {
			logger.error("[PdfSignatureFactory.sign]::No se encuentra el alias '" + alias + "' en el dispositivo", e);
			throw new SignatureException(i18n.getString("error.no.alias.dispositivo", alias), e);
		} catch (HashingException e) {
			logger.error("[PdfSignatureFactory.sign]::Error obteniendo el hash del documento", e);
			throw new SignatureException(i18n.getString("error.hash.documento"), e);
		} catch (LoadingObjectException e) {
			logger.error("[PdfSignatureFactory.sign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[PdfSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} catch (PDFDocumentException e) {
			logger.error("[PdfSignatureFactory.sign]::El documento no es un PDF o es un PDF mal formado", e);
			throw new SignatureException(i18n.getString("error.pdf.mal.formado"), e);
		} catch (CertificateCANotFoundException e) {
			logger.error("[PdfSignatureFactory.sign]::El certificado no pertenece a una de las autoridades de certificacin de confianza", e);
			throw new SignatureException(i18n.getString("error.no.ca.confianza"), e);
		} catch (InvalidCertificateException e) {
			logger.error("[PdfSignatureFactory.sign]::El certificado con el que se pretende firmar est revocado", e);
			throw new SignatureException(i18n.getString("error.firmando.certificado.revocado"), e);
		} 
		
		logger.debug("[PdfSignatureFactory.sign]::Se ha obtenido la firma");
		try {
			return Util.encodeBase64(signature.toByteArray());
		} catch (IOException e) {
			logger.error("[PdfSignatureFactory.sign]::No se ha podido leer el fichero PDF firmado", e);
			throw new SignatureException(i18n.getString("error.lectura.pdf.firmado"), e);
		}
	}

	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una cofirma a partir de la firma y el documento en fichero
	 * 
	 * @param fSignature Fichero con la firma
	 * @param fDocument Documento que origin la firma (puede ser nulo)
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 * @throws SignatureNotFoundException El fichero de la firma no existe o no puede ser
	 * 	ledo
	 */
	public String coSign(File fSignature, File fDocument, DeviceManager manager, String alias, String reference) throws SignatureException, SignatureNotFoundException {
		logger.info("[Pkcs7SignatureFactory.coSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}

	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una cofirma a partir de los contenidos de la firma y el documento
	 * 
	 * @param signatureContent Contenido de la firma
	 * @param documentContent Contenido del documento que origin la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 */
	public String coSign(byte[] signatureContent, byte[] documentContent, DeviceManager manager, String alias, String reference) throws SignatureException {
		logger.info("[Pkcs7SignatureFactory.coSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}

	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param signature Firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return
	 * @throws SignatureException
	 */
	public String counterSign(byte[] signature, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException {
		logger.info("[PdfSignatureFactory.counterSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}
	
	/**
	 * MTODO NO IMPLEMENTADO!!! <BR><BR>
	 * 
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param signatureFile Fichero con la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return
	 * @throws SignatureException
	 */
	public String counterSign(File signatureFile, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException {
		logger.info("[PdfSignatureFactory.counterSign]::Mtodo no implementado en la librera Arang");
		throw new SignatureException (i18n.getString("error.metodo.no.implementado"));
	}
	
	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un fichero.
	 * 
	 * @param signature Fichero con la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(File signature) throws Exception {
		return new PDFSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un array de bytes.
	 * 
	 * @param signature Contenido de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(byte[] signature) throws Exception {
		return new PDFSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un stream de lectura.
	 * 
	 * @param isSignature Stream al fichero de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(InputStream isSignature) throws Exception {
		return new PDFSignature(isSignature);
	}

}
