package es.accv.arangi.client.signature;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.document.ByteArrayDocument;
import es.accv.arangi.base.document.FileDocument;
import es.accv.arangi.base.document.IDocument;
import es.accv.arangi.base.exception.device.LoadingObjectException;
import es.accv.arangi.base.exception.document.HashingException;
import es.accv.arangi.base.exception.document.InitDocumentException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.exception.signature.XMLDocumentException;
import es.accv.arangi.base.signature.ISignature;
import es.accv.arangi.base.signature.XAdESBESSignature;
import es.accv.arangi.base.signature.util.XAdESDataObjectFormat;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.client.util.I18n;

/**
 * Clase que realiza firmas XAdES-BES detached (al estilo @Firma).
 * La firma es un XAdES-BES attached donde el documento es el hash sha1 en base64
 * del documento original.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class XadesBesDetachedAFirmaSignatureFactory extends XadesBesSignatureFactory {

	/*
	 * Path a la plantilla
	 */
	private static final String TEMPLATE_XADES_AFIRMA = "es/accv/arangi/client/template/arangi-xades-afirma.xml";

	/*
	 * Nombre del tag documento en la plantilla
	 */
	private static final String TAG_DOCUMENT_ID = "CONTENT";

	/*
	 * Nombre del tag padre de la firma
	 */
	private static final String TAG_SIGNATURE_PARENT = "AFIRMA";

	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(XadesBesDetachedAFirmaSignatureFactory.class);
	
	/*
	 * i18n
	 */
	I18n i18n;
	
	/**
	 * Constructor de todas las factory de firma
	 * 
	 * @param attached La firma es attached
	 * @param i18n Internacionalizacin
	 */
	public XadesBesDetachedAFirmaSignatureFactory (Boolean attached, I18n i18n) {
		this.i18n = i18n;
	}

	/* (non-Javadoc)
	 * @see es.accv.arangi.client.signature.ArangiClientSignatureFactory#sign(java.io.File, es.accv.arangi.base.device.DeviceManager, java.lang.String)
	 */
	public String sign(File file, DeviceManager manager, String alias, String reference) throws NoDocumentToSignException, SignatureException {
		logger.info("[XadesBesDetachedAFirmaSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { file, manager, alias, reference }));
		
		//-- Obtener el documento
		FileDocument document;
		try {
			document = new FileDocument(file);
		} catch (InitDocumentException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::No ha sido posible leer el fichero " + file, e);
			throw new NoDocumentToSignException(i18n.getString("error.no.documento.leer", file), e);
		}
		
		//-- Obtener el XML
		ByteArrayDocument xml = new ByteArrayDocument (getXML(document));
		logger.debug("Obtenido el XML a firmar");
		
		//-- Si existe obtener el objeto con la descripcin del documento a firmar
		XAdESDataObjectFormat dof = getDofFromOptions();
		
		//-- Firmar
		logger.debug("[XadesBesDetachedAFirmaSignatureFactory.sign]::Firmando...");
		XAdESBESSignature signature;
		try {
			signature = XAdESBESSignature.signAttached(manager, alias, xml, TAG_DOCUMENT_ID, TAG_SIGNATURE_PARENT, dof);
		} catch (LoadingObjectException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} catch (XMLDocumentException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::Error montando el fichero XML", e);
			throw new SignatureException(i18n.getString("error.montando.xml"), e);
		} 
		
		logger.debug("[XadesBesDetachedAFirmaSignatureFactory.sign]::Se ha obtenido la firma");
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}

	public String sign(byte[] contentToSign, DeviceManager manager, String alias, String reference) throws SignatureException {
		logger.info("[XadesBesDetachedAFirmaSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { contentToSign, manager, alias, reference }));
		
		//-- Obtener el documento
		ByteArrayDocument document = new ByteArrayDocument(contentToSign);
		
		//-- Obtener el XML
		ByteArrayDocument xml = new ByteArrayDocument (getXML(document));
		logger.debug("Obtenido el XML a firmar");
		
		//-- Si existe obtener el objeto con la descripcin del documento a firmar
		XAdESDataObjectFormat dof = getDofFromOptions();
		
		//-- Firmar
		logger.debug("[XadesBesDetachedAFirmaSignatureFactory.sign]::Firmando...");
		XAdESBESSignature signature;
		try {
			signature = XAdESBESSignature.signAttached(manager, alias, xml, TAG_DOCUMENT_ID, TAG_SIGNATURE_PARENT, dof);
		} catch (LoadingObjectException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} catch (XMLDocumentException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::Error montando el fichero XML", e);
			throw new SignatureException(i18n.getString("error.montando.xml"), e);
		} 
		
		logger.debug("[XadesBesDetachedAFirmaSignatureFactory.sign]::Se ha obtenido la firma");
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.sign]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un fichero.
	 * 
	 * @param signature Fichero con la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(File signature) throws Exception {
		return new XAdESBESSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un array de bytes.
	 * 
	 * @param signature Contenido de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(byte[] signature) throws Exception {
		return new XAdESBESSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un stream de lectura.
	 * 
	 * @param isSignature Stream al fichero de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(InputStream isSignature) throws Exception {
		return new XAdESBESSignature(isSignature);
	}

	//-- Mtodos privados
	
	/*
	 * Obtiene el XML a firmar. En el tag del documento ir el hash sha1 del fichero
	 * en base64.
	 */
	private byte[] getXML(IDocument document) throws SignatureException {
		
		String hash;
		try {
			hash = Util.encodeBase64(document.getHash());
		} catch (HashingException e) {
			logger.info("[XadesBesDetachedAFirmaSignatureFactory.getXML]::No es posible obtener el hash del documento", e);
			throw new SignatureException("No es posible obtener el hash del documento", e);
		}
		logger.debug ("Obtenido el hash del documento: " + hash);
		
		//-- Obtener la plantilla
		DocumentBuilder db;
		Document doc;
		try {
			db = getDocumentBuilder ();
			doc = db.parse(new Util().getClass().getClassLoader().getResourceAsStream(TEMPLATE_XADES_AFIRMA));
		} catch (SAXException e) {
			logger.info("[XadesBesDetachedAFirmaSignatureFactory.getXML]::No ha sido posible obtener la plantilla de Arang-XAdES", e);
			throw new SignatureException ("No ha sido posible obtener la plantilla de Arang-XAdES", e);
		} catch (IOException e) {
			logger.info("[XadesBesDetachedAFirmaSignatureFactory.getXML]::No ha sido posible obtener la plantilla de Arang-XAdES", e);
			throw new SignatureException ("No ha sido posible obtener la plantilla de Arang-XAdES", e);
		} catch (XMLDocumentException e) {
			logger.info("[XadesBesDetachedAFirmaSignatureFactory.getXML]::No ha sido posible obtener el DocumentBuilder para parsear el fichero XML", e);
			throw new SignatureException ("No ha sido posible obtener el DocumentBuilder para parsear el fichero XML", e);
		}
		
		//-- Buscar el elemento documento
		XPathFactory factory = XPathFactory.newInstance();
		XPath xpath = factory.newXPath();
		NodeList signatureNodes;
		try {
			XPathExpression expr = xpath.compile("//*[@Id='" + TAG_DOCUMENT_ID + "']");
			signatureNodes = (NodeList) expr.evaluate (doc, XPathConstants.NODESET);
		} catch (XPathExpressionException e) {
			logger.info("[XadesBesDetachedAFirmaSignatureFactory.getXML]::No ha sido posible obtener el elemento documento dentro de la plantilla", e);
			throw new SignatureException ("No ha sido posible obtener el elemento documento dentro de la plantilla", e);
		}
		Element eDocument = (Element) signatureNodes.item(0);
		eDocument.appendChild(doc.createTextNode(hash));

		//-- Devolver el XML
		try {
	        TransformerFactory tf = TransformerFactory.newInstance();
	    	Transformer trans = tf.newTransformer();
	    	
	    	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	    	trans.transform(new DOMSource(doc), new StreamResult(baos));
			return baos.toByteArray();
		} catch (TransformerConfigurationException e) {
			//-- no se va a dar un problema con las transformaciones, ya habra dado en la entrada
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.getXML]::Error en las transformaciones", e);
			throw new SignatureException ("Error en las transformaciones XML", e);
		} catch (TransformerException e) {
			//-- no se va a dar un problema con las transformaciones, ya habra dado en la entrada
			logger.error("[XadesBesDetachedAFirmaSignatureFactory.getXML]::Error en las transformaciones", e);
			throw new SignatureException ("Error en las transformaciones XML", e);
		} 
	}

	   /*
     * Obtiene el document builder para generar documentos DOM
     */
	protected DocumentBuilder getDocumentBuilder() throws XMLDocumentException {

		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
    	dbf.setNamespaceAware(true);

		DocumentBuilder db;
		try {
			db = dbf.newDocumentBuilder();
		} catch (ParserConfigurationException e1) {
			logger.info ("[XadesBesDetachedAFirmaSignatureFactory.getDocumentBuilder]::No se puede inicializar el documento XML vaco", e1);
			throw new XMLDocumentException("No se puede inicializar el documento XML vaco", e1);
		}
		
		return db;
	}


}
