package es.accv.arangi.client.signature;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;

import org.apache.log4j.Logger;

import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.document.ByteArrayDocument;
import es.accv.arangi.base.document.FileDocument;
import es.accv.arangi.base.exception.device.LoadingObjectException;
import es.accv.arangi.base.exception.document.InitDocumentException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.signature.ISignature;
import es.accv.arangi.base.signature.XAdESBESSignature;
import es.accv.arangi.base.signature.util.XAdESDataObjectFormat;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.client.util.I18n;

/**
 * Clase que realiza firmas XAdES-BES detached
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public class XadesBesDetachedSignatureFactory extends XadesBesSignatureFactory {

	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(XadesBesDetachedSignatureFactory.class);
	
	/*
	 * i18n
	 */
	I18n i18n;
	
	/**
	 * Constructor de todas las factory de firma
	 * 
	 * @param attached La firma es attached
	 * @param i18n Internacionalizacin
	 */
	public XadesBesDetachedSignatureFactory (Boolean attached, I18n i18n) {
		this.i18n = i18n;
	}

	/* (non-Javadoc)
	 * @see es.accv.arangi.client.signature.ArangiClientSignatureFactory#sign(java.io.File, es.accv.arangi.base.device.DeviceManager, java.lang.String)
	 */
	public String sign(File file, DeviceManager manager, String alias, String reference) throws NoDocumentToSignException, SignatureException {
		logger.info("[XadesBesDetachedSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { file, manager, alias, reference }));
		
		//-- Si existe obtener el objeto con la descripcin del documento a firmar
		XAdESDataObjectFormat dof = getDofFromOptions();
		
		//-- Firmar
		logger.debug("[XadesBesDetachedSignatureFactory.sign]::Firmando...");
		XAdESBESSignature signature;
		try {
			if (reference == null) {
				//-- La referencia ser el path del fichero
				signature = XAdESBESSignature.signDetached(manager, alias, file, dof);
			} else {
				//-- Obtener el documento
				FileDocument document;
				try {
					document = new FileDocument(file);
				} catch (InitDocumentException e) {
					logger.error("[XadesBesDetachedSignatureFactory.sign]::No ha sido posible leer el fichero " + file, e);
					throw new NoDocumentToSignException(i18n.getString("error.no.documento.leer", file), e);
				}
				signature = XAdESBESSignature.signDetached(manager, alias, document, reference, dof);
			}
		} catch (LoadingObjectException e) {
			logger.error("[XadesBesDetachedSignatureFactory.sign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[XadesBesDetachedSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} catch (NoDocumentToSignException e) {
			logger.error("[XadesBesDetachedSignatureFactory.sign]::El fichero a firmar es nulo o no existe", e);
			throw e;
		}
		
		logger.debug("[XadesBesDetachedSignatureFactory.sign]::Se ha obtenido la firma");
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesBesDetachedSignatureFactory.sign]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}

	/* (non-Javadoc)
	 * @see es.accv.arangi.client.signature.ArangiClientSignatureFactory#sign(java.io.File, es.accv.arangi.base.device.DeviceManager, java.lang.String)
	 */
	public String sign(byte[] contentToSign, DeviceManager manager, String alias, String reference) throws SignatureException {
		logger.info("[XadesBesDetachedSignatureFactory.sign]::Entrada::" + Arrays.asList(new Object[] { contentToSign, manager, alias, reference }));
		
		//-- Obtener el documento
		ByteArrayDocument document = new ByteArrayDocument(contentToSign);
		
		//-- Si existe obtener el objeto con la descripcin del documento a firmar
		XAdESDataObjectFormat dof = getDofFromOptions();
		
		//-- Firmar
		logger.debug("[XadesBesDetachedSignatureFactory.sign]::Firmando...");
		XAdESBESSignature signature;
		try {
			signature = XAdESBESSignature.signDetached(manager, alias, document, reference, dof);
		} catch (LoadingObjectException e) {
			logger.error("[XadesBesDetachedSignatureFactory.sign]::No es posible obtener la clave privada o el certificado para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[XadesBesDetachedSignatureFactory.sign]::Error durante el proceso de firma", e);
			throw new SignatureException(i18n.getString("error.firmando"), e);
		} 
		
		logger.debug("[XadesBesDetachedSignatureFactory.sign]::Se ha obtenido la firma");
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesBesDetachedSignatureFactory.sign]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un fichero.
	 * 
	 * @param signature Fichero con la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(File signature) throws Exception {
		return new XAdESBESSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un array de bytes.
	 * 
	 * @param signature Contenido de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(byte[] signature) throws Exception {
		return new XAdESBESSignature(signature);
	}

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un stream de lectura.
	 * 
	 * @param isSignature Stream al fichero de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	public ISignature getSignatureObject(InputStream isSignature) throws Exception {
		return new XAdESBESSignature(isSignature);
	}


}
