package es.accv.arangi.client.signature;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;

import org.apache.log4j.Logger;

import es.accv.arangi.base.certificate.Certificate;
import es.accv.arangi.base.device.DeviceManager;
import es.accv.arangi.base.document.ByteArrayDocument;
import es.accv.arangi.base.document.FileDocument;
import es.accv.arangi.base.exception.device.LoadingObjectException;
import es.accv.arangi.base.exception.document.HashingException;
import es.accv.arangi.base.exception.document.InitDocumentException;
import es.accv.arangi.base.exception.signature.NoCoincidentDocumentException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.exception.signature.SignatureNotFoundException;
import es.accv.arangi.base.exception.signature.XMLDocumentException;
import es.accv.arangi.base.signature.XAdESBESSignature;
import es.accv.arangi.base.util.Util;
import es.accv.arangi.client.SignatureConstants;
import es.accv.arangi.client.util.I18n;

/**
 * Clase base de todas las firmas XAdES-BES de este paquete.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public abstract class XadesBesSignatureFactory extends BaseSignatureFactory {

	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(XadesBesSignatureFactory.class);
	
	/*
	 * i18n
	 */
	protected I18n i18n;
	
	/**
	 * Realiza una cofirma a partir de la firma y el documento en fichero
	 * 
	 * @param fSignature Fichero con la firma
	 * @param fDocument Documento que origin la firma (puede ser nulo)
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 * @throws SignatureNotFoundException El fichero de la firma no existe o no puede ser
	 * 	ledo
	 */
	public String coSign(File fSignature, File fDocument, DeviceManager manager, String alias, String reference) throws SignatureException, SignatureNotFoundException {
		logger.info("[XadesBesSignatureFactory.coSign]::Entrada::" + Arrays.asList(new Object[] { fSignature, fDocument, manager, alias, reference }));
		
		//-- Obtener el documento
		FileDocument document = null;
		try {
			document = new FileDocument(fDocument);
		} catch (InitDocumentException e) {
		}
		
		//-- Obtener la firma
		logger.debug("[XadesBesSignatureFactory.coSign]::Obteniendo la firma");
		XAdESBESSignature signature;
		try {
			signature = new XAdESBESSignature(fSignature);
		} catch (XMLDocumentException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::El fichero no parece un XML vlido");
			throw new SignatureException(i18n.getString("error.firma.no.xml"), e);
		} catch (FileNotFoundException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No existe el fichero con la firma " + fSignature, e);
			throw new SignatureNotFoundException(i18n.getString("error.no.firma", fDocument), e);
		}
		
		//-- Cofirmar
		logger.debug("[XadesBesSignatureFactory.coSign]::Cofirmando...");
		try {
			signature.coSign(manager, alias, document);
		} catch (SignatureNotFoundException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No existe ninguna firma en el fichero de firma " + fSignature, e);
			throw new SignatureException(i18n.getString("error.no.firma.en.firma"), e);
		} catch (NoDocumentToSignException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No existe el documento a cofirmar " + fDocument, e);
			throw new SignatureException(i18n.getString("error.no.documento.firmar"), e);
		} catch (HashingException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::Error realizando el hash del documento " + fDocument, e);
			throw new SignatureException(i18n.getString("error.hash"), e);
		} catch (LoadingObjectException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::Error durante el proceso de cofirma", e);
			throw new SignatureException(i18n.getString("error.cofirmando"), e);
		} catch (NoCoincidentDocumentException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::El documento a cofirmar no se corresponde con el documento de la primera firma", e);
			throw new SignatureException(i18n.getString("error.documentos.no.coinciden"), e);
		}
		
		logger.debug("[XadesBesSignatureFactory.coSign]::Se ha realizado la cofirma");
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}

	/**
	 * Realiza una cofirma a partir de los contenidos de la firma y el documento
	 * 
	 * @param signatureContent Contenido de la firma
	 * @param documentContent Contenido del documento que origin la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para firmar 
	 * @param reference Referencia a la que apuntar la firma en caso detached. Para la 
	 * 	firma en PDFs se utiliza como motivo de la firma.
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de cofirma
	 */
	public String coSign(byte[] signatureContent, byte[] documentContent, DeviceManager manager, String alias, String reference) throws SignatureException {
		logger.info("[XadesBesSignatureFactory.coSign]::Entrada::" + Arrays.asList(new Object[] { signatureContent, documentContent, manager, alias, reference }));
		
		//-- Obtener el documento
		ByteArrayDocument document = null;
		if (documentContent != null) {
			document = new ByteArrayDocument(documentContent);
		}
		
		//-- Obtener la firma
		logger.debug("[XadesBesSignatureFactory.coSign]::Obteniendo la firma");
		XAdESBESSignature signature;
		try {
			signature = new XAdESBESSignature(signatureContent);
		} catch (XMLDocumentException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::El fichero no parece un XML vlido");
			throw new SignatureException(i18n.getString("error.firma.no.xml"), e);
		} 
		
		//-- Cofirmar
		logger.debug("[XadesBesSignatureFactory.coSign]::Cofirmando...");
		try {
			signature.coSign(manager, alias, document);
		} catch (SignatureNotFoundException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No existe ninguna firma en el fichero de firma ", e);
			throw new SignatureException(i18n.getString("error.no.firma.en.firma"), e);
		} catch (NoDocumentToSignException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No existe el documento a cofirmar", e);
			throw new SignatureException(i18n.getString("error.no.documento.firmar"), e);
		} catch (HashingException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::Error realizando el hash del documento", e);
			throw new SignatureException(i18n.getString("error.hash"), e);
		} catch (LoadingObjectException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No es posible obtener la clave privada para firmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (SignatureException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::Error durante el proceso de cofirma", e);
			throw new SignatureException(i18n.getString("error.cofirmando"), e);
		} catch (NoCoincidentDocumentException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::El documento a cofirmar no se corresponde con el documento de la primera firma", e);
			throw new SignatureException(i18n.getString("error.documentos.no.coinciden"), e);
		}
		
		logger.debug("[XadesBesSignatureFactory.coSign]::Se ha realizado la cofirma");
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesBesSignatureFactory.coSign]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}

	/**
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param signatureFile Fichero con la firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de contrafirma
	 */
	public String counterSign(File signatureFile, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException {
		logger.info("[XadesSignatureFactory.counterSignBes]::Entrada::" + Arrays.asList(new Object[] { signatureFile, manager, alias, signatureToCounterSignCertificate }));
		
		//-- Cargar el objeto firma
		XAdESBESSignature signature;
		try {
			signature = new XAdESBESSignature(signatureFile);
			logger.debug("[XadesSignatureFactory.counterSignBes]::Se ha cargado el objeto firma");
		} catch (XMLDocumentException e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No se ha podido obtener el objeto firma", e);
			throw new SignatureException(i18n.getString("error.cargando.firma", SignatureConstants.SIGNATURE_XADES_BES_DETACHED), e);
		} catch (FileNotFoundException e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No ha sido posible leer la firma en fichero " + signatureFile, e);
			throw new SignatureException(i18n.getString("error.no.documento.leer", signatureFile), e);
		}
		
		//-- Contrafirmar
		try {
			signature.counterSign(manager, alias, signatureToCounterSignCertificate);
			logger.debug("[XadesSignatureFactory.counterSignBes]::Se ha realizado la contrafirma");
		} catch (LoadingObjectException e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No es posible obtener la clave privada para contrafirmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (Exception e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No se ha podido realizar la contrafirma", e);
			throw new SignatureException(i18n.getString("error.contrafirmando"), e);
		}
		
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}

	/**
	 * Realiza una contrafirma para la firma cuyo certificado se pasa en el parmetro 
	 * 'signatureToCounterSignCertificate'. Si este parmetro es nulo se contrafirmar
	 * la ltima firma.
	 * 
	 * @param bSignature Firma
	 * @param manager Dispositivo
	 * @param alias Alias dentro del dispositivo del certificado para contrafirmar 
	 * @param signatureToCounterSignCertificate Certificado de la firma que se 
	 * 	contrafirmar (puede ser nulo).
	 * @return Firma en base64
	 * @throws SignatureException Error durante el proceso de contrafirma
	 */
	public String counterSign(byte[] bSignature, DeviceManager manager, String alias, 
			Certificate signatureToCounterSignCertificate) throws SignatureException {
		logger.info("[XadesSignatureFactory.counterSignBes]::Entrada::" + Arrays.asList(new Object[] { bSignature, manager, alias, signatureToCounterSignCertificate }));
		
		//-- Cargar el objeto firma
		XAdESBESSignature signature;
		try {
			signature = new XAdESBESSignature(bSignature);
			logger.debug("[XadesSignatureFactory.counterSignBes]::Se ha cargado el objeto firma");
		} catch (XMLDocumentException e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No se ha podido obtener el objeto firma", e);
			throw new SignatureException(i18n.getString("error.cargando.firma", SignatureConstants.SIGNATURE_XADES_BES_DETACHED), e);
		}
		
		//-- Contrafirmar
		try {
			signature.counterSign(manager, alias, signatureToCounterSignCertificate);
			logger.debug("[XadesSignatureFactory.counterSignBes]::Se ha realizado la contrafirma");
		} catch (LoadingObjectException e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No es posible obtener la clave privada para contrafirmar", e);
			throw new SignatureException(i18n.getString("error.no.clave.privada"), e);
		} catch (Exception e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No se ha podido realizar la contrafirma", e);
			throw new SignatureException(i18n.getString("error.contrafirmando"), e);
		}
		
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			signature.save(baos);
			return Util.encodeBase64(baos.toByteArray());
		} catch (IOException e) {
			logger.error("[XadesSignatureFactory.counterSignBes]::No se ha podido leer la firma", e);
			throw new SignatureException(i18n.getString("error.lectura.firma"), e);
		}
	}
	
}
