/**
 * 
 */
package es.accv.arangi.client.validation;

import java.io.File;
import java.io.InputStream;
import java.util.Arrays;

import org.apache.log4j.Logger;

import es.accv.arangi.base.document.ByteArrayDocument;
import es.accv.arangi.base.document.IDocument;
import es.accv.arangi.base.document.InputStreamDocument;
import es.accv.arangi.base.exception.certificate.NormalizeCertificateException;
import es.accv.arangi.base.exception.document.HashingException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;
import es.accv.arangi.base.signature.ISignature;
import es.accv.arangi.base.util.validation.ValidationResult;
import es.accv.arangi.client.exception.SignatureFormatException;
import es.accv.arangi.client.exception.ValidationException;
import es.accv.arangi.client.model.ValidationObject;
import es.accv.arangi.client.util.I18n;
import es.accv.arangi.util.ArangiUtil;


/**
 * Interfaz que han de definir las clases que validan los diferentes tipos
 * de firma del cliente de Arang.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public abstract class ArangiClientValidationFactory {
	
	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(ArangiClientValidationFactory.class);
	
	/*
	 * i18n
	 */
	I18n i18n;
	
	/**
	 * Constructor de todas las factory de validacin
	 * 
	 * @param i18n Internacionalizacin
	 */
	public ArangiClientValidationFactory (I18n i18n) {
		this.i18n = i18n;
	}

	
	public ValidationObject validate(File signature) throws SignatureFormatException, ValidationException {
		logger.debug ("Entrada::" + signature);
		
		//-- Obtener el objeto para la validacin
		ISignature signatureObject;
		try {
			signatureObject = getSignatureObject(signature);
		} catch (Exception e) {
			logger.error("No se ha podido obtener el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
		
		//-- Validar
		try {
			return getValidationResult (signatureObject);
		} catch (Exception e) {
			logger.error("No se ha podido validar el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new ValidationException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
	}

	public ValidationObject validate(byte[] signature) throws SignatureFormatException, ValidationException {
		logger.debug ("Entrada::" + signature);
		
		//-- Obtener el objeto para la validacin
		ISignature signatureObject;
		try {
			signatureObject = getSignatureObject(signature);
		} catch (Exception e) {
			logger.error("No se ha podido obtener el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
		
		//-- Validar
		try {
			return getValidationResult (signatureObject);
		} catch (Exception e) {
			logger.error("No se ha podido validar el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new ValidationException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
	}

	public ValidationObject validate(InputStream isSignature) throws SignatureFormatException, ValidationException {
		logger.debug ("Entrada::" + isSignature);
		
		//-- Obtener el objeto para la validacin
		ISignature signatureObject;
		try {
			signatureObject = getSignatureObject(isSignature);
		} catch (Exception e) {
			logger.error("No se ha podido obtener el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
		
		//-- Validar
		try {
			return getValidationResult (signatureObject);
		} catch (Exception e) {
			logger.error("No se ha podido validar el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new ValidationException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
	}

	public ValidationObject validate(File signature, ByteArrayDocument document) throws SignatureFormatException, ValidationException {
		logger.debug ("Entrada::" + signature);
		
		//-- Obtener el objeto para la validacin
		ISignature signatureObject;
		try {
			signatureObject = getSignatureObject(signature);
		} catch (Exception e) {
			logger.error("No se ha podido obtener el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
		
		//-- Validar
		try {
			return getValidationResult (signatureObject, document);
		} catch (Exception e) {
			logger.error("No se ha podido validar el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new ValidationException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
	}

	public ValidationObject validate(byte[] signature, ByteArrayDocument document) throws SignatureFormatException, ValidationException {
		logger.debug ("Entrada::" + signature);
		
		//-- Obtener el objeto para la validacin
		ISignature signatureObject;
		try {
			signatureObject = getSignatureObject(signature);
		} catch (Exception e) {
			logger.error("No se ha podido obtener el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
		
		//-- Validar
		try {
			return getValidationResult (signatureObject, document);
		} catch (Exception e) {
			logger.error("No se ha podido validar el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new ValidationException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
	}

	public ValidationObject validate(InputStream isSignature, InputStreamDocument document) throws SignatureFormatException, ValidationException {
		logger.debug ("Entrada::" + isSignature);
		
		//-- Obtener el objeto para la validacin
		ISignature signatureObject;
		try {
			signatureObject = getSignatureObject(isSignature);
		} catch (Exception e) {
			logger.error("No se ha podido obtener el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new SignatureFormatException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
		
		//-- Validar
		try {
			return getValidationResult (signatureObject, document);
		} catch (Exception e) {
			logger.error("No se ha podido validar el objeto firma del tipo '" + getSignatureType() + "'", e);
			throw new ValidationException(i18n.getString("error.cargando.firma", getSignatureType()), e);
		}
	}
	
	//-- Mtodos abstractos

	/** 
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un fichero.
	 * 
	 * @param signature Fichero con la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	protected abstract ISignature getSignatureObject(File signature) throws Exception;

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un array de bytes.
	 * 
	 * @param signature Contenido de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	protected abstract ISignature getSignatureObject(byte[] signature) throws Exception;

	/**
	 * Obtiene un objeto para tratar la firma adecuada inicializado con un stream de lectura.
	 * 
	 * @param isSignature Stream al fichero de la firma
	 * @return Objeto para tratar la firma
	 * @throws Exception Error cargando la firma
	 */
	protected abstract ISignature getSignatureObject(InputStream isSignature) throws Exception;

	/**
	 * Obtiene en texto el tipo de firma que se est tratando
	 * 
	 * @return tipo de firma
	 */
	protected abstract String getSignatureType();

	//-- Mtodos privados
	
	/*
	 * Obtiene el resultado de la validacin para una firma attached
	 */
	private ValidationObject getValidationResult(ISignature signatureObject) throws Exception {
		logger.info("[ArangiClient.getValidationResult]::Entrada::" + signatureObject);
		
		//-- Validar
		ValidationResult[] validationResults;
		try {
			validationResults = signatureObject.isValid(ArangiUtil.getACCVCaList());
			logger.info("[ArangiClient.getValidationResult]::Se ha validado la firma");
		} catch (HashingException e) {
			logger.error("[ArangiClient.getValidationResult]::Error obteniendo el hash del documento", e);
			throw new Exception(i18n.getString("error.hash.documento"), e);
		} catch (SignatureException e) {
			logger.error("[ArangiClient.getValidationResult]::Error tratando el objeto firma", e);
			throw new Exception(i18n.getString("error.tratando.firma"), e);
		} catch (NormalizeCertificateException e) {
			logger.error("[ArangiClient.getValidationResult]::Error normalizando el certificado", e);
			throw new Exception(i18n.getString("error.normalizando.certificado.firma"), e);
		} catch (NoDocumentToSignException e) {
			logger.error("[ArangiClient.getValidationResult]::La firma es detached y por tanto no hay documento", e);
			throw new Exception(i18n.getString("error.validar.attached"), e);
		}
		
		//-- Construir el objeto resultado
		return new ValidationObject (signatureObject.getSignatureType(), validationResults, i18n);
	}

	/*
	 * Obtiene el resultado de la validacin para una firma dettached
	 */
	private ValidationObject getValidationResult(ISignature signatureObject, IDocument document) throws Exception {
		logger.info("[ArangiClient.getValidationResult]::Entrada::" + Arrays.asList(new Object[] { signatureObject, document }));
		
		//-- Validar
		ValidationResult[] validationResults;
		try {
			validationResults = signatureObject.isValid(document, ArangiUtil.getACCVCaList());
			logger.info("[ArangiClient.getValidationResult]::Se ha validado la firma");
		} catch (HashingException e) {
			logger.error("[ArangiClient.getValidationResult]::Error obteniendo el hash del documento", e);
			throw new Exception(i18n.getString("error.hash.documento"), e);
		} catch (SignatureException e) {
			logger.error("[ArangiClient.getValidationResult]::Error tratando el objeto firma", e);
			throw new Exception(i18n.getString("error.tratando.firma"), e);
		} catch (NormalizeCertificateException e) {
			logger.error("[ArangiClient.getValidationResult]::Error normalizando el certificado", e);
			throw new Exception(i18n.getString("error.normalizando.certificado.firma"), e);
		} 
		
		//-- Construir el objeto resultado
		return new ValidationObject (signatureObject.getSignatureType(), validationResults, i18n);
	}



}
